# import dependencies
from pyomo.environ import ConcreteModel, Set, Var, Objective, minimize, Constraint, log
from pyomo.core.expr.numvalue import NumericValue
import numpy as np
import pandas as pd
from ..constant import CET_ADDI, CET_MULT, FUN_PROD, FUN_COST, RTS_CRS, RTS_VRS, OPT_DEFAULT, OPT_LOCAL
from .tools import optimize_model


class CNLSG1:
    """initial Group-VC-added CNLS (CNLS+G) model
    """

    def __init__(self, y, x, cutactive, cet=CET_ADDI, fun=FUN_PROD, rts=RTS_VRS):
        """CNLS+G model 1

        Args:
            y (float): output variable.
            x (float): input variables.
            cutactive (float): active concavity constraint.
            cet (String, optional): CET_ADDI (additive composite error term) or CET_MULT (multiplicative composite error term). Defaults to CET_ADDI.
            fun (String, optional): FUN_PROD (production frontier) or FUN_COST (cost frontier). Defaults to FUN_PROD.
            rts (String, optional): RTS_VRS (variable returns to scale) or RTS_CRS (constant returns to scale). Defaults to RTS_VRS.
        """
        # TODO(error/warning handling): Check the configuration of the model exist
        self.x, self.y, self.cet, self.fun, self.rts = x, y, cet, fun, rts
        self.cutactive = cutactive

        # Initialize the CNLS model
        self.__model__ = ConcreteModel()

        # Initialize the sets
        self.__model__.I = Set(initialize=range(len(self.y)))
        self.__model__.J = Set(initialize=range(len(self.x[0])))

        # Initialize the variables
        self.__model__.alpha = Var(self.__model__.I, doc='alpha')
        self.__model__.beta = Var(self.__model__.I,
                                  self.__model__.J,
                                  bounds=(0.0, None),
                                  doc='beta')
        self.__model__.epsilon = Var(self.__model__.I, doc='residual')
        self.__model__.frontier = Var(self.__model__.I,
                                      bounds=(0.0, None),
                                      doc='estimated frontier')

        # Setup the objective function and constraints
        self.__model__.objective = Objective(rule=self.__objective_rule(),
                                             sense=minimize,
                                             doc='objective function')
        self.__model__.regression_rule = Constraint(self.__model__.I,
                                                    rule=self.__regression_rule(),
                                                    doc='regression equation')
        if self.cet == CET_MULT:
            self.__model__.log_rule = Constraint(self.__model__.I,
                                                 rule=self.__log_rule(),
                                                 doc='log-transformed regression equation')
        self.__model__.afriat_rule = Constraint(self.__model__.I,
                                                rule=self.__afriat_rule(),
                                                doc='elementary Afriat approach')
        self.__model__.sweet_rule = Constraint(self.__model__.I,
                                               self.__model__.I,
                                               rule=self.__sweet_rule(),
                                               doc='sweet spot approach')

        # Optimize model
        self.optimization_status, self.problem_status = 0, 0

    def optimize(self, email=OPT_LOCAL, solver=OPT_DEFAULT):
        """Optimize the function by requested method

        Args:
            email (string): The email address for remote optimization. It will optimize locally if OPT_LOCAL is given.
            solver (string): The solver chosen for optimization. It will optimize with default solver if OPT_DEFAULT is given.
        """
        # TODO(error/warning handling): Check problem status after optimization
        self.problem_status, self.optimization_status = optimize_model(
            self.__model__, email, self.cet, solver)

    def __objective_rule(self):
        """Return the proper objective function"""

        def objective_rule(model):
            return sum(model.epsilon[i] ** 2 for i in model.I)

        return objective_rule

    def __regression_rule(self):
        """Return the proper regression constraint"""
        if self.cet == CET_ADDI:
            if self.rts == RTS_VRS:
                def regression_rule(model, i):
                    return self.y[i] == model.alpha[i] \
                        + sum(model.beta[i, j] * self.x[i][j] for j in model.J) \
                        + model.epsilon[i]

                return regression_rule
            elif self.rts == RTS_CRS:
                def regression_rule(model, i):
                    return self.y[i] == sum(model.beta[i, j] * self.x[i][j] for j in model.J) \
                        + model.epsilon[i]

                return regression_rule
        elif self.cet == CET_MULT:

            def regression_rule(model, i):
                return log(
                    self.y[i]) == log(model.frontier[i] + 1) + model.epsilon[i]

            return regression_rule

        raise ValueError("Undefined model parameters.")

    def __log_rule(self):
        """Return the proper log constraint"""
        if self.cet == CET_MULT:
            if self.rts == RTS_VRS:

                def log_rule(model, i):
                    return model.frontier[i] == model.alpha[i] + sum(
                        model.beta[i, j] * self.x[i][j] for j in model.J) - 1

                return log_rule
            elif self.rts == RTS_CRS:

                def log_rule(model, i):
                    return model.frontier[i] == sum(
                        model.beta[i, j] * self.x[i][j] for j in model.J) - 1

                return log_rule

        raise ValueError("Undefined model parameters.")

    def __afriat_rule(self):
        """Return the proper elementary Afriat approach constraint"""
        if self.fun == FUN_PROD:
            __operator = NumericValue.__le__
        elif self.fun == FUN_COST:
            __operator = NumericValue.__ge__

        if self.cet == CET_ADDI:
            if self.rts == RTS_VRS:

                def afriat_rule(model, i):
                    return __operator(
                        model.alpha[i] + sum(model.beta[i, j] * self.x[i][j]
                                             for j in model.J),
                        model.alpha[self.__model__.I.nextw(i)] +
                        sum(model.beta[self.__model__.I.nextw(i), j] * self.x[i][j]
                            for j in model.J))

                return afriat_rule
            elif self.rts == RTS_CRS:

                def afriat_rule(model, i):
                    return __operator(sum(model.beta[i, j] * self.x[i][j]
                                          for j in model.J),
                                      sum(model.beta[self.__model__.I.nextw(i), j] * self.x[i][j]
                                          for j in model.J))

                return afriat_rule
        elif self.cet == CET_MULT:
            if self.rts == RTS_VRS:

                def afriat_rule(model, i):
                    return __operator(
                        model.alpha[i] + sum(model.beta[i, j] * self.x[i][j]
                                             for j in model.J),
                        model.alpha[self.__model__.I.nextw(i)] +
                        sum(model.beta[self.__model__.I.nextw(i), j] * self.x[i][j]
                            for j in model.J))

                return afriat_rule
            elif self.rts == RTS_CRS:

                def afriat_rule(model, i):
                    return __operator(
                        sum(model.beta[i, j] * self.x[i][j] for j in model.J),
                        sum(model.beta[self.__model__.I.nextw(i), j] * self.x[i][j] for j in model.J))

                return afriat_rule

        raise ValueError("Undefined model parameters.")

    def __sweet_rule(self, ):
        """Return the proper sweet spot approach constraint"""
        if self.fun == FUN_PROD:
            __operator = NumericValue.__le__
        elif self.fun == FUN_COST:
            __operator = NumericValue.__ge__

        if self.cet == CET_ADDI:
            if self.rts == RTS_VRS:

                def sweet_rule(model, i, h):
                    if self.cutactive[i][h]:
                        if i == h:
                            return Constraint.Skip
                        return __operator(model.alpha[i] + sum(model.beta[i, j] * self.x[i][j]
                                                               for j in model.J),
                                          model.alpha[h] + sum(model.beta[h, j] * self.x[i][j]
                                                               for j in model.J))
                    return Constraint.Skip

                return sweet_rule
            elif self.rts == RTS_CRS:

                def sweet_rule(model, i, h):
                    if self.cutactive[i][h]:
                        if i == h:
                            return Constraint.Skip
                        return __operator(sum(model.beta[i, j] * self.x[i][j]
                                              for j in model.J),
                                          sum(model.beta[h, j] * self.x[i][j]
                                              for j in model.J))
                    return Constraint.Skip

                return sweet_rule
        elif self.cet == CET_MULT:
            if self.rts == RTS_VRS:

                def sweet_rule(model, i, h):
                    if self.cutactive[i][h]:
                        if i == h:
                            return Constraint.Skip
                        return __operator(model.alpha[i] + sum(model.beta[i, j] * self.x[i][j]
                                                               for j in model.J),
                                          model.alpha[h] + sum(model.beta[h, j] * self.x[i][j]
                                                               for j in model.J))
                    return Constraint.Skip

                return sweet_rule
            elif self.rts == RTS_CRS:

                def sweet_rule(model, i, h):
                    if self.cutactive[i][h]:
                        if i == h:
                            return Constraint.Skip
                        return __operator(sum(model.beta[i, j] * self.x[i][j] for j in model.J),
                                          sum(model.beta[h, j] * self.x[i][j] for j in model.J))
                    return Constraint.Skip

                return sweet_rule

        raise ValueError("Undefined model parameters.")

    def get_alpha(self):
        """Return alpha value by array"""
        if self.optimization_status == 0:
            self.optimize()
        alpha = list(self.__model__.alpha[:].value)
        return np.asarray(alpha)

    def get_beta(self):
        """Return beta value by array"""
        if self.optimization_status == 0:
            self.optimize()
        beta = np.asarray([i + tuple([j]) for i, j in zip(list(self.__model__.beta),
                                                          list(self.__model__.beta[:, :].value))])
        beta = pd.DataFrame(beta, columns=['Name', 'Key', 'Value'])
        beta = beta.pivot(index='Name', columns='Key', values='Value')
        return beta.to_numpy()



# import dependencies
from pyomo.environ import Objective, minimize
from . import CQER
from .constant import CET_ADDI, FUN_PROD, RTS_VRS
from .utils import tools


class wCQR(CQER.CQR):
    """Weighted Convex Quantile Regression (wCQR)
    """

    def __init__(self, y, x, w, tau, z=None, cet=CET_ADDI, fun=FUN_PROD, rts=RTS_VRS):
        """wCQR model

        Args:
            y (float): output variable. 
            x (float): input variables.
            w (float): weight variable.
            z (float, optional): Contextual variable(s). Defaults to None.
            tau (float): quantile.
            cet (String, optional): CET_ADDI (additive composite error term) or CET_MULT (multiplicative composite error term). Defaults to CET_ADDI.
            fun (String, optional): FUN_PROD (production frontier) or FUN_COST (cost frontier). Defaults to FUN_PROD.
            rts (String, optional): RTS_VRS (variable returns to scale) or RTS_CRS (constant returns to scale). Defaults to RTS_VRS.
        """
        # TODO(error/warning handling): Check the configuration of the model exist
        super().__init__(y, x, tau, z, cet, fun, rts)
        self.w = tools.trans_list(tools.to_1d_list(w))

        self.__model__.objective.deactivate()
        self.__model__.weighted_objective = Objective(
            rule=self.__weighted_objective_rule(), sense=minimize, doc='weighted objective rule')

    def __weighted_objective_rule(self):
        def weighted_objective_rule(model):
            return self.tau * sum(self.w[i] * model.epsilon_plus[i] for i in model.I) \
                + (1 - self.tau) * \
                sum(self.w[i] * model.epsilon_minus[i] for i in model.I)

        return weighted_objective_rule


class wCER(CQER.CQR):
    """Weighted Convex Expectile Regression (wCER)
    """

    def __init__(self, y, x, w, tau, z=None, cet=CET_ADDI, fun=FUN_PROD, rts=RTS_VRS):
        """wCER model

        Args:
            y (float): output variable. 
            x (float): input variables.
            w (float): weight variable.
            z (float, optional): Contextual variable(s). Defaults to None.
            tau (float): expectile.
            cet (String, optional): CET_ADDI (additive composite error term) or CET_MULT (multiplicative composite error term). Defaults to CET_ADDI.
            fun (String, optional): FUN_PROD (production frontier) or FUN_COST (cost frontier). Defaults to FUN_PROD.
            rts (String, optional): RTS_VRS (variable returns to scale) or RTS_CRS (constant returns to scale). Defaults to RTS_VRS.
        """
        super().__init__(y, x, tau, z, cet, fun, rts)
        self.w = tools.trans_list(tools.to_1d_list(w))

        self.__model__.objective.deactivate()
        self.__model__.weighted_squared_objective = Objective(
            rule=self.__weighted_squared_objective_rule(), sense=minimize, doc='weighted square objective rule')

    def __weighted_squared_objective_rule(self):
        def weighted_squared_objective_rule(model):
            return self.tau * sum(self.w[i] * model.epsilon_plus[i] ** 2 for i in model.I) \
                + (1 - self.tau) * \
                sum(self.w[i] * model.epsilon_minus[i] ** 2 for i in model.I)

        return weighted_squared_objective_rule



# import dependencies
from pyomo.environ import Objective, minimize
from . import pCNLS
from .constant import CET_ADDI, FUN_PROD, RTS_VRS
from .utils import tools


class pwCNLS(pCNLS.pCNLS):
    """penalized Weighted Convex Nonparametric Least Square (pwCNLS)
    """

    def __init__(self, y, x, w, eta, z=None, cet=CET_ADDI, fun=FUN_PROD, rts=RTS_VRS, penalty=1):
        """wCNLS model

        Args:
            y (float): output variable. 
            x (float): input variables.
            w (float): weight variable.
            eta (float): regularization parameter.
            z (float, optional): Contextual variable(s). Defaults to None.
            cet (String, optional): CET_ADDI (additive composite error term) or CET_MULT (multiplicative composite error term). Defaults to CET_ADDI.
            fun (String, optional): FUN_PROD (production frontier) or FUN_COST (cost frontier). Defaults to FUN_PROD.
            rts (String, optional): RTS_VRS (variable returns to scale) or RTS_CRS (constant returns to scale). Defaults to RTS_VRS.
            penalty (int, optional): penalty=1 (L1 norm) and penalty=2 (L2 norm). Defaults to 1.
        """
        # TODO(error/warning handling): Check the configuration of the model exist
        self.w = tools.trans_list(tools.to_1d_list(w))
        pCNLS.pCNLS.__init__(self, y, x, eta, z, cet, fun, rts, penalty)

        self.__model__.objective.deactivate()
        self.__model__.new_objective = Objective(rule=self.__new_objective_rule(),
                                                 sense=minimize,
                                                 doc='weighted objective function')

    def __new_objective_rule(self):
        """Return the proper objective function"""

        def objective_rule(model):
            return sum(self.w[i] * model.epsilon[i] ** 2 for i in model.I)

        return objective_rule



# import dependencies
import numpy as np
from scipy.spatial.distance import cdist
from .tools import trans_list, to_2d_list


def sweet(x):
    """Sweet spot approach

    Args:
        x (float): input variables.

    Returns:
        list: active concavity constraint.
    """

    # transform data
    df = np.asmatrix(to_2d_list(trans_list(x)))

    # calculate distance matrix
    distance = cdist(df, df)
    distance[np.diag_indices_from(distance)] = np.nan

    # calculate distance cut
    distcut = np.asmatrix(np.nanpercentile(distance, 3, axis=0))

    # find concavity constraint in sweet spot
    distance = np.where(np.isnan(distance), 0, distance)
    cutactive = np.zeros((len(distance), len(distance)))
    for i in range(len(distance)):
        for j in range(len(distance)):
            if distance[i, j] <= distcut[:, i]:
                cutactive[i, j] = 1

    return to_2d_list(trans_list(cutactive))



# import dependencies
from pyomo.environ import ConcreteModel, Set, Var, Objective, minimize, Constraint, log
from pyomo.core.expr.numvalue import NumericValue
import numpy as np
import pandas as pd

from . import CNLS
from constant import CET_MULT, RDF_DI, RDF_DO, OPT_DEFAULT, RTS_CRS, RTS_VRS, OPT_LOCAL
from .utils import tools


class CNLSRDF(CNLS.CNLS):
    """Convex Nonparametric Least Square with radial distance function
    """

    def __init__(self, y, x, z=None, rdf=RDF_DI, rts=RTS_VRS):
        """CNLSRDF model
        Args:
            y (float): output variable. 
            x (float): input variables.
            z (float, optional): control variables. Defaults to None.
            cet (String, optional): CET_ADDI (additive composite error term) or CET_MULT (multiplicative composite error term). Defaults to CET_ADDI.
            rdf (String, optional): RDF_DI (input distance function) or RDF_DO (output distance function). Defaults to RDF_DI.
            rts (String, optional): RTS_VRS (variable returns to scale) or RTS_CRS (constant returns to scale). Defaults to RTS_VRS.
        """
        # TODO(error/warning handling): Check the configuration of the model exist
        self.y, self.x, self.z = tools.assert_valid_mupltiple_x_y_data(y, x, z)
        self.rdf, self.rts = rdf, rts

        # rescale the input/output variables
        if self.rdf == RDF_DI:
            self.X = np.array(self.x)
            self.X = np.cumprod(self.X**(1/self.X.shape[1]), axis=1)[:, -1]
            self.X = tools.to_1d_list(tools.trans_list(self.X))

        elif self.rdf == RDF_DO:
            self.Y = np.array(self.y)
            self.Y = np.cumprod(self.Y**(1/self.Y.shape[1]), axis=1)[:, -1]
            self.Y = tools.to_1d_list(tools.trans_list(self.Y))

        # Initialize the CNLS model
        self.__model__ = ConcreteModel()

        if type(self.z) != type(None):
            # Initialize the set of z
            self.__model__.K = Set(initialize=range(len(self.z[0])))

            # Initialize the variables for z variable
            self.__model__.lamda = Var(self.__model__.K, doc='z coefficient')

        # Initialize the sets
        self.__model__.I = Set(initialize=range(len(self.y)))
        self.__model__.J = Set(initialize=range(len(self.x[0])))
        self.__model__.Q = Set(initialize=range(len(self.y[0])))
        if self.rdf == RDF_DI:
            self.__model__.JJ = Set(initialize=range(1, len(self.x[0])))
        elif self.rdf == RDF_DO:
            self.__model__.QQ = Set(initialize=range(1, len(self.y[0])))

        # Initialize the variables
        self.__model__.alpha = Var(self.__model__.I, doc='alpha')
        self.__model__.beta = Var(self.__model__.I,
                                  self.__model__.J,
                                  bounds=(0.0, None),
                                  doc='beta')
        self.__model__.epsilon = Var(self.__model__.I, doc='residual')
        self.__model__.chi = Var(self.__model__.I,
                                 bounds=(0.0, None),
                                 doc='estimated radial distance function')
        self.__model__.gamma = Var(self.__model__.I,
                                   self.__model__.Q,
                                   bounds=(0.0, None),
                                   doc='gamma')

        if self.rdf == RDF_DI:
            self.__model__.kappa = Var(self.__model__.JJ, doc='free parameter')
        elif self.rdf == RDF_DO:
            self.__model__.kappa = Var(self.__model__.QQ, doc='free parameter')

        # Setup the objective function and constraints
        self.__model__.objective = Objective(rule=self._CNLS__objective_rule(),
                                             sense=minimize,
                                             doc='objective function')
        self.__model__.regression_rule = Constraint(self.__model__.I,
                                                    rule=self.__regression_rule(),
                                                    doc='regression equation')
        self.__model__.log_rule = Constraint(self.__model__.I,
                                             rule=self.__log_rule(),
                                             doc='log-transformed regression equation')
        self.__model__.afriat_rule = Constraint(self.__model__.I,
                                                self.__model__.I,
                                                rule=self.__afriat_rule(),
                                                doc='afriat inequality')

        # Optimize model
        self.optimization_status, self.problem_status = 0, 0

    def optimize(self, email=OPT_LOCAL, solver=OPT_DEFAULT):
        """Optimize the function by requested method
        Args:
            email (string): The email address for remote optimization. It will optimize locally if OPT_LOCAL is given.
            solver (string): The solver chosen for optimization. It will optimize with default solver if OPT_DEFAULT is given.
        """
        # TODO(error/warning handling): Check problem status after optimization
        self.problem_status, self.optimization_status = tools.optimize_model(
            self.__model__, email, CET_MULT, solver)

    def __regression_rule(self):
        """Return the proper regression constraint"""
        if self.rdf == RDF_DI:

            if type(self.z) != type(None):
                def regression_rule(model, i):
                    return log(self.x[i][0]) == -log(model.chi[i] + 1) + sum(model.kappa[j] *
                                                                             (log(
                                                                                 self.x[i][0]) - log(self.x[i][j]))
                                                                             for j in model.JJ) + sum(model.lamda[k] * self.z[i][k]
                                                                                                      for k in model.K) + model.epsilon[i]

                return regression_rule

            def regression_rule(model, i):
                return log(self.x[i][0]) == -log(model.chi[i] + 1) + sum(model.kappa[j] *
                                                                         (log(
                                                                             self.x[i][0]) - log(self.x[i][j]))
                                                                         for j in model.JJ) + model.epsilon[i]

            return regression_rule
        elif self.rdf == RDF_DO:

            if type(self.z) != type(None):
                def regression_rule(model, i):
                    return log(self.y[i][0]) == -log(model.chi[i] + 1) + sum(model.kappa[q] *
                                                                             (log(
                                                                                 self.y[i][0]) - log(self.y[i][q]))
                                                                             for q in model.QQ) + sum(model.lamda[k] * self.z[i][k]
                                                                                                      for k in model.K) + model.epsilon[i]

                return regression_rule

            def regression_rule(model, i):
                return log(self.y[i][0]) == -log(model.chi[i] + 1) + sum(model.kappa[q] *
                                                                         (log(
                                                                             self.y[i][0]) - log(self.y[i][q]))
                                                                         for q in model.QQ) + model.epsilon[i]

            return regression_rule

        raise ValueError("Undefined model parameters.")

    def __log_rule(self):
        """Return the proper log constraint"""
        if self.rdf == RDF_DI:
            if self.rts == RTS_VRS:

                def log_rule(model, i):
                    return model.chi[i] == model.alpha[i] + sum(
                        model.beta[i, j] * (self.x[i][j]/self.X[i]) for j in model.J) \
                        - sum(model.gamma[i, q] * self.y[i][q]
                              for q in model.Q) - 1

                return log_rule
            elif self.rts == RTS_CRS:

                def log_rule(model, i):
                    return model.chi[i] == sum(
                        model.beta[i, j] * (self.x[i][j]/self.X[i]) for j in model.J) \
                        - sum(model.gamma[i, q] * self.y[i][q]
                              for q in model.Q) - 1

                return log_rule

        elif self.rdf == RDF_DO:
            if self.rts == RTS_VRS:

                def log_rule(model, i):
                    return model.chi[i] == model.alpha[i] + sum(
                        model.gamma[i, q] * (self.y[i][q]/self.Y[i]) for q in model.Q) \
                        - sum(model.beta[i, j] * self.x[i][j]
                              for j in model.J) - 1

                return log_rule
            elif self.rts == RTS_CRS:

                def log_rule(model, i):
                    return model.chi[i] == sum(
                        model.gamma[i, q] * (self.y[i][q]/self.Y[i]) for q in model.Q) \
                        - sum(model.beta[i, j] * self.x[i][j]
                              for j in model.J) - 1

                return log_rule

        raise ValueError("Undefined model parameters.")

    def __afriat_rule(self):
        """Return the proper afriat inequality constraint"""
        if self.rdf == RDF_DI:
            __operator = NumericValue.__le__
        elif self.rdf == RDF_DO:
            __operator = NumericValue.__ge__

        if self.rdf == RDF_DI:
            if self.rts == RTS_VRS:

                def afriat_rule(model, i, h):
                    if i == h:
                        return Constraint.Skip
                    return __operator(
                        model.alpha[i] + sum(model.beta[i, j] * (self.x[i][j]/self.X[i])
                                             for j in model.J) - sum(model.gamma[i, q] * self.y[i][q]
                                                                     for q in model.Q),
                        model.alpha[h] + sum(model.beta[h, j] * (self.x[i][j]/self.X[i])
                                             for j in model.J) - sum(model.gamma[h, q] * self.y[i][q]
                                                                     for q in model.Q))

                return afriat_rule
            elif self.rts == RTS_CRS:

                def afriat_rule(model, i, h):
                    if i == h:
                        return Constraint.Skip
                    return __operator(
                        sum(model.beta[i, j] * (self.x[i][j]/self.X[i])
                            for j in model.J) - sum(model.gamma[i, q] * self.y[i][q]
                                                    for q in model.Q),
                        sum(model.beta[h, j] * (self.x[i][j]/self.X[i])
                            for j in model.J) - sum(model.gamma[h, q] * self.y[i][q]
                                                    for q in model.Q))

                return afriat_rule

        elif self.rdf == RDF_DO:
            if self.rts == RTS_VRS:

                def afriat_rule(model, i, h):
                    if i == h:
                        return Constraint.Skip
                    return __operator(
                        model.alpha[i] + sum(model.gamma[i, q] * (self.y[i][q]/self.Y[i])
                                             for q in model.Q) - sum(model.beta[i, j] * self.x[i][j]
                                                                     for j in model.J),
                        model.alpha[h] + sum(model.gamma[h, q] * (self.y[i][q]/self.Y[i])
                                             for q in model.Q) - sum(model.beta[h, j] * self.x[i][j]
                                                                     for j in model.J))

                return afriat_rule
            elif self.rts == RTS_CRS:

                def afriat_rule(model, i, h):
                    if i == h:
                        return Constraint.Skip
                    return __operator(
                        sum(model.gamma[i, q] * (self.y[i][q]/self.Y[i])
                            for q in model.Q) - sum(model.beta[i, j] * self.x[i][j]
                                                    for j in model.J),
                        sum(model.gamma[h, q] * (self.y[i][q]/self.Y[i])
                            for q in model.Q) - sum(model.beta[h, j] * self.x[i][j]
                                                    for j in model.J))

                return afriat_rule

        raise ValueError("Undefined model parameters.")

    def display_gamma(self):
        """Display gamma value"""
        tools.assert_optimized(self.optimization_status)
        self.__model__.gamma.display()

    def get_gamma(self):
        """Return gamma value by array"""
        tools.assert_optimized(self.optimization_status)
        gamma = np.asarray([i + tuple([j]) for i, j in zip(list(self.__model__.gamma),
                                                           list(self.__model__.gamma[:, :].value))])
        gamma = pd.DataFrame(gamma, columns=['Name', 'Key', 'Value'])
        gamma = gamma.pivot(index='Name', columns='Key', values='Value')
        return gamma.to_numpy()

    def display_kappa(self):
        """Display kappa value"""
        tools.assert_optimized(self.optimization_status)
        self.__model__.kappa.display()

    def get_kappa(self):
        """Return kappa value by array"""
        tools.assert_optimized(self.optimization_status)
        kappa = list(self.__model__.kappa[:].value)
        return np.asarray(kappa)

    def get_chi(self):
        """Return estimated radial distance function by array"""
        tools.assert_optimized(self.optimization_status)
        chi = np.asarray(list(self.__model__.chi[:].value)) + 1
        return np.asarray(chi)



# import dependencies
from pyomo.environ import ConcreteModel, Set, Var, Objective, minimize, Constraint, log
from pyomo.core.expr.numvalue import NumericValue
import numpy as np
import pandas as pd

from . import CNLS
from .constant import CET_ADDI, CET_MULT, FUN_PROD, FUN_COST, OPT_DEFAULT, RTS_CRS, RTS_VRS, OPT_LOCAL
from .utils import tools


class weakCNLS(CNLS.CNLS):
    """Convex Nonparametric Least Square with weak disposability (weakCNLS)
    """

    def __init__(self, y, x, b, z=None, cet=CET_ADDI, fun=FUN_PROD, rts=RTS_VRS):
        """weakCNLS model

        Args:
            y (float): output variable. 
            x (float): input variables.
            b (float): undersiable variables.
            z (float, optional): Contextual variable(s). Defaults to None.
            cet (String, optional): CET_ADDI (additive composite error term) or CET_MULT (multiplicative composite error term). Defaults to CET_ADDI.
            fun (String, optional): FUN_PROD (production frontier) or FUN_COST (cost frontier). Defaults to FUN_PROD.
            rts (String, optional): RTS_VRS (variable returns to scale) or RTS_CRS (constant returns to scale). Defaults to RTS_VRS.
        """
        # TODO(error/warning handling): Check the configuration of the model exist
        self.y, self.x, self.b, self.z = tools.assert_valid_wp_data(y, x, b, z)
        self.cet, self.fun, self.rts = cet, fun, rts

        # Initialize the CNLS model
        self.__model__ = ConcreteModel()

        if type(self.z) != type(None):
            # Initialize the set of z
            self.__model__.K = Set(initialize=range(len(self.z[0])))

            # Initialize the variables for z variable
            self.__model__.lamda = Var(self.__model__.K, doc='z coefficient')

        # Initialize the sets
        self.__model__.I = Set(initialize=range(len(self.y)))
        self.__model__.J = Set(initialize=range(len(self.x[0])))
        self.__model__.L = Set(initialize=range(len(self.b[0])))

        # Initialize the variables
        self.__model__.alpha = Var(self.__model__.I, doc='alpha')
        self.__model__.beta = Var(self.__model__.I,
                                  self.__model__.J,
                                  bounds=(0.0, None),
                                  doc='beta')
        self.__model__.delta = Var(self.__model__.I,
                                   self.__model__.L,
                                   bounds=(0.0, None),
                                   doc='delta')
        self.__model__.epsilon = Var(self.__model__.I, doc='residual')
        self.__model__.frontier = Var(self.__model__.I,
                                      bounds=(0.0, None),
                                      doc='estimated frontier')

        # Setup the objective function and constraints
        self.__model__.objective = Objective(rule=self._CNLS__objective_rule(),
                                             sense=minimize,
                                             doc='objective function')
        self.__model__.regression_rule = Constraint(self.__model__.I,
                                                    rule=self.__regression_rule(),
                                                    doc='regression equation')
        if self.cet == CET_MULT:
            self.__model__.log_rule = Constraint(self.__model__.I,
                                                 rule=self.__log_rule(),
                                                 doc='log-transformed regression equation')
        self.__model__.afriat_rule = Constraint(self.__model__.I,
                                                self.__model__.I,
                                                rule=self.__afriat_rule(),
                                                doc='afriat inequality')
        self.__model__.disposability_rule = Constraint(self.__model__.I,
                                                       self.__model__.I,
                                                       rule=self.__disposability_rule(),
                                                       doc='weak disposibility')

        # Optimize model
        self.optimization_status, self.problem_status = 0, 0

    def optimize(self, email=OPT_LOCAL, solver=OPT_DEFAULT):
        """Optimize the function by requested method

        Args:
            email (string): The email address for remote optimization. It will optimize locally if OPT_LOCAL is given.
            solver (string): The solver chosen for optimization. It will optimize with default solver if OPT_DEFAULT is given.
        """
        # TODO(error/warning handling): Check problem status after optimization
        self.problem_status, self.optimization_status = tools.optimize_model(
            self.__model__, email, self.cet, solver)

    def __regression_rule(self):
        """Return the proper regression constraint"""
        if self.cet == CET_ADDI:
            if self.rts == RTS_VRS:
                if type(self.z) != type(None):
                    def regression_rule(model, i):
                        return self.y[i] == model.alpha[i] \
                            + sum(model.beta[i, j] * self.x[i][j] for j in model.J) \
                            + sum(model.delta[i, l] * self.b[i][l] for l in model.L) \
                            + sum(model.lamda[k] * self.z[i][k] for k in model.K) \
                            + model.epsilon[i]

                    return regression_rule

                def regression_rule(model, i):
                    return self.y[i] == model.alpha[i] \
                        + sum(model.beta[i, j] * self.x[i][j] for j in model.J) \
                        + sum(model.delta[i, l] * self.b[i][l] for l in model.L) \
                        + model.epsilon[i]

                return regression_rule
            elif self.rts == RTS_CRS:
                if type(self.z) != type(None):
                    def regression_rule(model, i):
                        return self.y[i] == sum(model.beta[i, j] * self.x[i][j] for j in model.J) \
                            + sum(model.delta[i, l] * self.b[i][l] for l in model.L) \
                            + sum(model.lamda[k] * self.z[i][k]
                                  for k in model.K) + model.epsilon[i]

                    return regression_rule

                def regression_rule(model, i):
                    return self.y[i] == sum(model.beta[i, j] * self.x[i][j] for j in model.J) \
                        + sum(model.delta[i, l] * self.b[i][l] for l in model.L) \
                        + model.epsilon[i]

                return regression_rule

        elif self.cet == CET_MULT:
            if type(self.z) != type(None):
                def regression_rule(model, i):
                    return log(self.y[i]) == log(model.frontier[i] + 1) \
                        + sum(model.lamda[k] * self.z[i][k]
                              for k in model.K) + model.epsilon[i]

                return regression_rule

            def regression_rule(model, i):
                return log(self.y[i]) == log(model.frontier[i] + 1) + model.epsilon[i]

            return regression_rule

        raise ValueError("Undefined model parameters.")

    def __log_rule(self):
        """Return the proper log constraint"""
        if self.cet == CET_MULT:
            if self.rts == RTS_VRS:

                def log_rule(model, i):
                    return model.frontier[i] == model.alpha[i] + sum(
                        model.beta[i, j] * self.x[i][j] for j in model.J) \
                        + sum(model.delta[i, l] * self.b[i][l]
                              for l in model.L) - 1

                return log_rule
            elif self.rts == RTS_CRS:

                def log_rule(model, i):
                    return model.frontier[i] == sum(
                        model.beta[i, j] * self.x[i][j] for j in model.J) \
                        + sum(model.delta[i, l] * self.b[i][l]
                              for l in model.L) - 1

                return log_rule

        raise ValueError("Undefined model parameters.")

    def __afriat_rule(self):
        """Return the proper afriat inequality constraint"""
        if self.fun == FUN_PROD:
            __operator = NumericValue.__le__
        elif self.fun == FUN_COST:
            __operator = NumericValue.__ge__

        if self.cet == CET_ADDI:
            if self.rts == RTS_VRS:

                def afriat_rule(model, i, h):
                    if i == h:
                        return Constraint.Skip
                    return __operator(
                        model.alpha[i] + sum(model.beta[i, j]
                                             * self.x[i][j] for j in model.J)
                        + sum(model.delta[i, l] * self.b[i][l]
                              for l in model.L),
                        model.alpha[h] + sum(model.beta[h, j]
                                             * self.x[i][j] for j in model.J)
                        + sum(model.delta[h, l] * self.b[i][l] for l in model.L))

                return afriat_rule
            elif self.rts == RTS_CRS:

                def afriat_rule(model, i, h):
                    if i == h:
                        return Constraint.Skip
                    return __operator(
                        sum(model.beta[i, j] * self.x[i][j] for j in model.J)
                        + sum(model.delta[i, l] * self.b[i][l]
                              for l in model.L),
                        sum(model.beta[h, j] * self.x[i][j] for j in model.J)
                        + sum(model.delta[h, l] * self.b[i][l] for l in model.L))

                return afriat_rule
        elif self.cet == CET_MULT:
            if self.rts == RTS_VRS:

                def afriat_rule(model, i, h):
                    if i == h:
                        return Constraint.Skip
                    return __operator(
                        model.alpha[i] + sum(model.beta[i, j]
                                             * self.x[i][j] for j in model.J)
                        + sum(model.delta[i, l] * self.b[i][l]
                              for l in model.L),
                        model.alpha[h] + sum(model.beta[h, j]
                                             * self.x[i][j] for j in model.J)
                        + sum(model.delta[h, l] * self.b[i][l] for l in model.L))

                return afriat_rule
            elif self.rts == RTS_CRS:

                def afriat_rule(model, i, h):
                    if i == h:
                        return Constraint.Skip
                    return __operator(
                        sum(model.beta[i, j] * self.x[i][j] for j in model.J)
                        + sum(model.delta[i, l] * self.b[i][l]
                              for l in model.L),
                        sum(model.beta[h, j] * self.x[i][j] for j in model.J)
                        + sum(model.delta[h, l] * self.b[i][l] for l in model.L))

                return afriat_rule

        raise ValueError("Undefined model parameters.")

    def __disposability_rule(self):
        """Return the proper weak disposability constraint"""
        def disposability_rule(model, i, h):
            if i == h:
                return Constraint.Skip
            return model.alpha[i] + sum(model.beta[i, j] * self.x[h][j] for j in model.J) >= 0

        return disposability_rule

    def display_delta(self):
        """Display delta value"""
        tools.assert_optimized(self.optimization_status)
        tools.assert_undesirable_output(self.b)
        self.__model__.delta.display()

    def get_delta(self):
        """Return delta value by array"""
        tools.assert_optimized(self.optimization_status)
        tools.assert_undesirable_output(self.b)
        delta = np.asarray([i + tuple([j]) for i, j in zip(list(self.__model__.delta),
                                                           list(self.__model__.delta[:, :].value))])
        delta = pd.DataFrame(delta, columns=['Name', 'Key', 'Value'])
        delta = delta.pivot(index='Name', columns='Key', values='Value')
        return delta.to_numpy()



# import dependencies
from pyomo.environ import ConcreteModel, Set, Var, Objective, minimize, Constraint
import numpy as np
import pandas as pd

from .constant import CET_ADDI, OPT_LOCAL, OPT_DEFAULT
from .utils import tools


class sCQR:
    """Simultaneous estimation of CQR
    """

    def __init__(self, y, x, tau, C):
        """sCQR model

        Args:
            y (float): output variable. 
            x (float): input variables.
            tau: vector of quantile.
            C: interval (small positive value)
        """
        # TODO(error/warning handling): Check the configuration of the model exist
        self.y, self.x = tools.to_1d_list(tools.trans_list(
            y)), tools.to_2d_list(tools.trans_list(x))
        self.tau = tools.to_1d_list(tools.trans_list(tau))
        self.C, self.cet = C, CET_ADDI

        # Initialize the CQR model
        self.__model__ = ConcreteModel()

        # Initialize the sets
        self.__model__.I = Set(initialize=range(len(self.y)))
        self.__model__.J = Set(initialize=range(len(self.x[0])))
        self.__model__.K = Set(initialize=range(len(self.tau)))
        self.__model__.W = Set(initialize=range(len(self.tau)-1))

        # Initialize the variables
        self.__model__.alpha = Var(
            self.__model__.K, self.__model__.I, doc='alpha')
        self.__model__.beta = Var(
            self.__model__.K, self.__model__.I, self.__model__.J, bounds=(0.0, None), doc='beta')
        self.__model__.epsilon_plus = Var(self.__model__.K, self.__model__.I, bounds=(
            0.0, None), doc='positive error term')
        self.__model__.epsilon_minus = Var(self.__model__.K, self.__model__.I, bounds=(
            0.0, None), doc='negative error term')

        # Setup the objective function and constraints
        self.__model__.objective = Objective(
            rule=self.__objective_rule(), sense=minimize, doc='objective function')

        self.__model__.regression_rule = Constraint(self.__model__.K, self.__model__.I, rule=self.__regression_rule(),
                                                    doc='regression equation')

        self.__model__.afriat_rule = Constraint(self.__model__.K, self.__model__.I, self.__model__.I, rule=self.__afriat_rule(),
                                                doc='afriat inequality')

        self.__model__.noncrossing_rule = Constraint(self.__model__.W, self.__model__.I, rule=self.__noncrossing_rule(),
                                                     doc='non-crossing constraint')

        # Optimize model
        self.optimization_status, self.problem_status = 0, 0

    def optimize(self, email=OPT_LOCAL, solver=OPT_DEFAULT):
        """Optimize the function by requested method

        Args:
            email (string): The email address for remote optimization. It will optimize locally if OPT_LOCAL is given.
            solver (string): The solver chosen for optimization. It will optimize with default solver if OPT_DEFAULT is given.
        """
        # TODO(error/warning handling): Check problem status after optimization
        self.problem_status, self.optimization_status = tools.optimize_model(
            self.__model__, email, self.cet, solver)

    def __objective_rule(self):
        """Return the proper objective function"""

        def objective_rule(model):
            return sum(self.tau[k] * sum(model.epsilon_plus[k, i] for i in model.I)
                       + (1 - self.tau[k]) * sum(model.epsilon_minus[k, i] for i in model.I) for k in model.K)

        return objective_rule

    def __regression_rule(self):
        """Return the proper regression constraint"""

        def regression_rule(model, k, i):
            return self.y[i] == model.alpha[k, i] + sum(model.beta[k, i, j] * self.x[i][j] for j in model.J) + \
                model.epsilon_plus[k, i] - model.epsilon_minus[k, i]

        return regression_rule

    def __afriat_rule(self):
        """Return the proper afriat inequality constraint"""

        def afriat_rule(model, k, i, h):
            if i == h:
                return Constraint.Skip
            return model.alpha[k, i] + sum(model.beta[k, i, j] * self.x[i][j] for j in model.J) \
                <= model.alpha[k, h] + sum(model.beta[k, h, j] * self.x[i][j] for j in model.J)

        return afriat_rule

    def __noncrossing_rule(self):
        """Return the proper non-crossing constraint"""

        def noncrossing_rule(model, w, i):
            return model.alpha[w, i] + sum(model.beta[w, i, j] * self.x[i][j] for j in model.J) + self.C <= \
                model.alpha[w+1, i] + \
                sum(model.beta[w+1, i, j] * self.x[i][j] for j in model.J)

        return noncrossing_rule

    def get_alpha(self):
        """Return alpha value by array"""
        tools.assert_optimized(self.optimization_status)
        alpha = np.asarray([i + tuple([j]) for i, j in zip(list(self.__model__.alpha),
                                                           list(self.__model__.alpha[:, :].value))])
        alpha = pd.DataFrame(alpha, columns=['Name', 'Key', 'Value'])
        alpha = alpha.pivot(index='Name', columns='Key', values='Value')
        return alpha.to_numpy().T

    def get_beta(self):
        """Return beta value by array"""
        tools.assert_optimized(self.optimization_status)
        beta = np.asarray([i + tuple([j]) for i, j in zip(list(self.__model__.beta),
                                                          list(self.__model__.beta[:, :, :].value))])
        beta = pd.DataFrame(beta, columns=['tau', 'n', 'd', 'beta'])
        betanew = []
        for i in range(len(self.tau)):
            beta_new = beta.loc[beta['tau'] == i]
            del beta_new["tau"]
            betanew.append(beta_new.pivot(
                index='n', columns='d', values='beta').to_numpy())
        return np.array(betanew).reshape(len(self.tau)*len(self.y), len(self.x[0]))

    def get_positive_residual(self):
        """Return positive residual value by array"""
        tools.assert_optimized(self.optimization_status)
        residual_plus = np.asarray([i + tuple([j]) for i, j in zip(list(self.__model__.epsilon_plus),
                                                                   list(self.__model__.epsilon_plus[:, :].value))])
        residual_plus = pd.DataFrame(residual_plus, columns=[
                                     'Name', 'Key', 'Value'])
        residual_plus = residual_plus.pivot(
            index='Name', columns='Key', values='Value')
        return residual_plus.to_numpy().T

    def get_negative_residual(self):
        """Return negative residual value by array"""
        tools.assert_optimized(self.optimization_status)
        residual_minus = np.asarray([i + tuple([j]) for i, j in zip(list(self.__model__.epsilon_minus),
                                                                    list(self.__model__.epsilon_minus[:, :].value))])
        residual_minus = pd.DataFrame(residual_minus, columns=[
                                      'Name', 'Key', 'Value'])
        residual_minus = residual_minus.pivot(
            index='Name', columns='Key', values='Value')
        return residual_minus.to_numpy().T

    def get_frontier(self):
        """Return estimated frontier value by array"""
        tools.assert_optimized(self.optimization_status)
        frontier = np.tile(np.asarray(self.y).reshape(len(self.y), 1), len(self.tau)) \
            - (self.get_positive_residual() - self.get_negative_residual())
        return np.asarray(frontier)


class sCER(sCQR):
    """Simultaneous estimation of CER
    """

    def __init__(self, y, x, tau, C):
        """sCER model

        Args:
            y (float): output variable. 
            x (float): input variables.
            tau: vector of expectile.
            C: interval (small positive value)
        """
        super().__init__(y, x, tau, C)
        self.__model__.objective.deactivate()
        self.__model__.squared_objective = Objective(
            rule=self.__squared_objective_rule(), sense=minimize, doc='squared objective rule')

    def __squared_objective_rule(self):
        def squared_objective_rule(model):
            return sum(self.tau[k] * sum(model.epsilon_plus[k, i] ** 2 for i in model.I)
                       + (1 - self.tau[k]) * sum(model.epsilon_minus[k, i] ** 2 for i in model.I) for k in model.K)

        return squared_objective_rule



# import dependencies
import matplotlib.pyplot as plt
import numpy as np
from .utils import interpolation
from .constant import FUN_PROD, FUN_COST, RED_MOM, RTS_VRS, RTS_CRS


def plot2d(model, x_select=0, label_name="estimated function", fig_name=None, method=RED_MOM):
    """Plot 2d estimated function/frontier

    Args:
        model: The input model for plotting.
        x_select (Integer): The selected x for plotting.
        label_name (String): the estimator name.
        fig_name (String, optional): The name of figure to save. Defaults to None.
    """
    x = np.array(model.x).T[x_select]
    y = np.array(model.y).T
    if y.ndim != 1:
        print("Plot with mutiple y is unavailable now.")
        return False
    if model.__class__.__name__ != "StoNED":
        yhat = np.array(model.get_frontier()).T
    else:
        yhat = np.array(model.get_frontier(method)).T

    data = (np.stack([x, y, yhat], axis=0)).T

    # sort
    data = data[np.argsort(data[:, 0])].T

    x, y, f = data[0], data[1], data[2]

    # create figure and axes objects
    fig, ax = plt.subplots()
    dp = ax.scatter(x, y, color="k", marker='x')
    fl = ax.plot(x, f, color="r", label=label_name)

    # add legend
    legend = plt.legend([dp, fl[0]],
                        ['Data points', label_name],
                        loc='upper left',
                        ncol=1,
                        fontsize=10,
                        frameon=False)

    # add x, y label
    ax.set_xlabel("Input $x$")
    ax.set_ylabel("Output $y$")

    # Remove top and right axes
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    if fig_name == None:
        plt.show()
    else:
        plt.savefig(fig_name)


def plot3d(model, x_select_1=0, x_select_2=1, fig_name=None, line_transparent=False, pane_transparent=False):
    """Plot 3d estimated function/frontier

    Args:
        model: The input model for plotting.
        x_select_1 (Integer): The selected x for plotting.
        x_select_2 (Integer): The selected x for plotting.
        fun (String, optional): FUN_PROD (production frontier) or FUN_COST (cost frontier). Defaults to FUN_PROD.       
        fig_name (String, optional): The name of figure to save. Defaults to None.
        line_transparent (bool, optional): control the transparency of the lines. Defaults to False.
        pane_transparent (bool, optional): control the transparency of the pane. Defaults to False.
    """
    x = np.array(model.x).T
    y = np.array(model.y).T
    if model.rts == RTS_VRS:
        alpha, beta = model.get_alpha(), model.get_beta()
    elif model.rts == RTS_CRS:
        alpha, beta = np.zeros((model.get_beta()).shape[0]), model.get_beta()
    if y.ndim != 1:
        print("Plot with mutiple y is unavailable now.")
        return False

    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    dp = ax.scatter(x[x_select_1], x[x_select_2], y, marker='.', s=10)

    # Revise the Z-axis left side
    tmp_planes = ax.zaxis._PLANES
    ax.zaxis._PLANES = (tmp_planes[2], tmp_planes[3],
                        tmp_planes[0], tmp_planes[1],
                        tmp_planes[4], tmp_planes[5])

    # make the grid lines transparent
    if line_transparent == False:
        ax.xaxis._axinfo["grid"]['color'] = (1, 1, 1, 0)
        ax.yaxis._axinfo["grid"]['color'] = (1, 1, 1, 0)
        ax.zaxis._axinfo["grid"]['color'] = (1, 1, 1, 0)

    # make the panes transparent
    if pane_transparent != False:
        ax.xaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        ax.yaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        ax.zaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))

    xlin1 = np.linspace(min(x[x_select_1]), max(x[x_select_1]), 30)
    xlin2 = np.linspace(min(x[x_select_2]), max(x[x_select_2]), 30)
    XX, YY = np.meshgrid(xlin1, xlin2)

    ZZ = np.zeros((len(XX), len(XX)))
    for i in range(len(XX)):
        for j in range(len(XX)):
            ZZ[i, j] = interpolation.interpolation(alpha, beta,
                                                   x=np.array(
                                                       [XX[i, j], YY[i, j]], ndmin=2),
                                                   fun=model.fun)

    ax.plot_surface(XX, YY, ZZ, rstride=1, cstride=1, cmap='viridis',
                    edgecolor='none', alpha=0.5)

    # add x, y, z label
    ax.set_xlabel("Input $x1$")
    ax.set_ylabel("Input $x2$")
    ax.set_zlabel("Output $y$", rotation=0)

    if fig_name == None:
        plt.show()
    else:
        plt.savefig(fig_name)



from . import CNLS
from . import CNLSDDF
from . import CNLSG
from . import CNLSRDF
from . import constant
from . import CQER
from . import CQERDDF
from . import CQERG
from . import CSVR
from . import dataset
from . import DEA
from . import FDH
from . import pCNLS
from . import pCQER
from . import pICQER
from . import pwCNLS
from . import pwCQER
from . import plot
from . import ICNLS
from . import ICQER
from . import sCQER
from . import StoNED
from . import wCNLS
from . import wCQER
from . import weakCNLS

__all__ = [
    'CNLS',
    'CNLSDDF',
    'CNLSG',
    'CNLSRDF',
    'constant',
    'CQER',
    'CQERDDF',
    'CQERG',
    'CSVR',
    'dataset',
    'DEA',
    'FDH',
    'pCNLS',
    'pCQER',
    'pICQER',
    'pwCNLS',
    'pwCQER',
    'plot',
    'ICNLS',
    'ICQER',
    'sCQER',
    'StoNED',
    'wCNLS',
    'wCQER',
    'weakCNLS',
]


# Composite error term
CET_ADDI = "addi"
"""
CET_ADDI: Additive composite error term.
"""

CET_MULT = "mult"
"""
CET_MULT: Multiplicative composite error term.
"""

CET_Categories = {
    CET_ADDI: "Additive composite error term",
    CET_MULT: "Multiplicative composite error term"
}

CET_Model_Categories = {
    CET_ADDI: "additive model",
    CET_MULT: "multiplicative model"
}

# Frontier
FUN_PROD = "prod"
"""
FUN_PROD: Production frontier.
"""

FUN_COST = "cost"
"""
FUN_COST: Cost frontier.
"""

FUN_Categories = {
    FUN_PROD: "Production frontier",
    FUN_COST: "Cost frontier"
}

# Return to scale
RTS_VRS = "vrs"
"""
RTS_VRS: Variable returns to scale.
"""

RTS_CRS = "crs"
"""
RTS_CRS: Constant returns to scale.
"""

RTS_Categories = {
    RTS_VRS: "Variable returns to scale",
    RTS_CRS: "Constant returns to scale"
}

# Orientation
ORIENT_IO = "io"
"""
ORIENT_IO: Input orientation.
"""

ORIENT_OO = "oo"
"""
ORIENT_OO: Output orientation.
"""

ORIENT_Categories = {
    ORIENT_IO: "Input orientation",
    ORIENT_OO: "Output orientation"
}

# Residual decomposition
RED_MOM = "MOM"
"""
RED_MOM: Method of moments.
"""

RED_QLE = "QLE"
"""
RED_QLE: Quassi-likelihood estimation.
"""

RED_KDE = "KDE"
"""
RED_KDE: Kernel deconvolution estimation.
"""

RED_Categories = {
    RED_MOM: "Method of moments",
    RED_QLE: "Quassi-likelihood estimation",
    RED_KDE: "Kernel deconvolution estimation"
}

# Optimization
OPT_LOCAL = "local"
OPT_DEFAULT = None

# Radial distance function
RDF_DI = "DI"
"""
RDF_DI: Input distance function.
"""

RDF_DO = "DO"
"""
RDF_DO: Output distance function.
"""

FUN_Categories = {
    RDF_DI: "Input distance function",
    RDF_DO: "Output distance function"
}


# import dependencies
from pyomo.environ import ConcreteModel, Set, Var, Objective, minimize, maximize, Constraint
import numpy as np
import pandas as pd
from .constant import CET_ADDI, ORIENT_IO, ORIENT_OO, RTS_VRS, RTS_CRS, OPT_DEFAULT, OPT_LOCAL
from .utils import tools


class DEA:
    """Data Envelopment Analysis (DEA)
    """

    def __init__(self, y, x, orient, rts, yref=None, xref=None):
        """DEA: Envelopment problem 

        Args:
            y (float): output variable. 
            x (float): input variables.
            orient (String): ORIENT_IO (input orientation) or ORIENT_OO (output orientation)
            rts (String): RTS_VRS (variable returns to scale) or RTS_CRS (constant returns to scale)
            yref (String, optional): reference output. Defaults to None.
            xref (String, optional): reference inputs. Defaults to None.
        """
        # TODO(error/warning handling): Check the configuration of the model exist
        self.y, self.x = tools.assert_valid_mupltiple_y_data(y, x)
        self.orient, self.rts = orient, rts

        if type(yref) != type(None):
            self.yref, self.xref = tools.assert_valid_reference_data(
                self.y, self.x, yref, xref)
        else:
            self.yref, self.xref = self.y, self.x

        # Initialize DEA model
        self.__model__ = ConcreteModel()
        self.__model__.R = Set(initialize=range(len(self.yref)))

        # Initialize sets
        self.__model__.I = Set(initialize=range(len(self.y)))
        self.__model__.J = Set(initialize=range(len(self.x[0])))
        self.__model__.K = Set(initialize=range(len(self.y[0])))

        # Initialize variable
        self.__model__.theta = Var(self.__model__.I, doc='efficiency')
        self.__model__.lamda = Var(self.__model__.I, self.__model__.R, bounds=(
            0.0, None), doc='intensity variables')

        # Setup the objective function and constraints
        if self.orient == ORIENT_IO:
            self.__model__.objective = Objective(
                rule=self.__objective_rule(), sense=minimize, doc='objective function')
        else:
            self.__model__.objective = Objective(
                rule=self.__objective_rule(), sense=maximize, doc='objective function')
        self.__model__.input = Constraint(
            self.__model__.I, self.__model__.J, rule=self.__input_rule(), doc='input constraint')
        self.__model__.output = Constraint(
            self.__model__.I, self.__model__.K, rule=self.__output_rule(), doc='output constraint')
        if self.rts == RTS_VRS:
            self.__model__.vrs = Constraint(
                self.__model__.I, rule=self.__vrs_rule(), doc='variable return to scale rule')

        # Optimize model
        self.optimization_status, self.problem_status = 0, 0

    def __objective_rule(self):
        """Return the proper objective function"""
        def objective_rule(model):
            return sum(model.theta[i] for i in model.I)
        return objective_rule

    def __input_rule(self):
        """Return the proper input constraint"""
        if self.orient == ORIENT_IO:
            def input_rule(model, o, j):
                return model.theta[o]*self.x[o][j] >= sum(model.lamda[o, r]*self.xref[r][j] for r in model.R)
            return input_rule
        elif self.orient == ORIENT_OO:
            def input_rule(model, o, j):
                return sum(model.lamda[o, r] * self.xref[r][j] for r in model.R) <= self.x[o][j]
            return input_rule

    def __output_rule(self):
        """Return the proper output constraint"""
        if self.orient == ORIENT_IO:
            def output_rule(model, o, k):
                return sum(model.lamda[o, r] * self.yref[r][k] for r in model.R) >= self.y[o][k]
            return output_rule
        elif self.orient == ORIENT_OO:
            def output_rule(model, o, k):
                return model.theta[o]*self.y[o][k] <= sum(model.lamda[o, r]*self.yref[r][k] for r in model.R)
            return output_rule

    def __vrs_rule(self):
        def vrs_rule(model, o):
            return sum(model.lamda[o, r] for r in model.R) == 1
        return vrs_rule

    def optimize(self, email=OPT_LOCAL, solver=OPT_DEFAULT):
        """Optimize the function by requested method

        Args:
            email (string): The email address for remote optimization. It will optimize locally if OPT_LOCAL is given.
            solver (string): The solver chosen for optimization. It will optimize with default solver if OPT_DEFAULT is given.
        """
        # TODO(error/warning handling): Check problem status after optimization
        self.problem_status, self.optimization_status = tools.optimize_model(
            self.__model__, email, CET_ADDI, solver)

    def display_status(self):
        """Display the status of problem"""
        print(self.optimization_status)

    def display_theta(self):
        """Display theta value"""
        tools.assert_optimized(self.optimization_status)
        self.__model__.theta.display()

    def display_lamda(self):
        """Display lamda value"""
        tools.assert_optimized(self.optimization_status)
        self.__model__.lamda.display()

    def get_status(self):
        """Return status"""
        return self.optimization_status

    def get_theta(self):
        """Return theta value by array"""
        tools.assert_optimized(self.optimization_status)
        theta = list(self.__model__.theta[:].value)
        return np.asarray(theta)

    def get_lamda(self):
        """Return lamda value by array"""
        tools.assert_optimized(self.optimization_status)
        lamda = np.asarray([i + tuple([j]) for i, j in zip(list(self.__model__.lamda),
                                                           list(self.__model__.lamda[:, :].value))])
        lamda = pd.DataFrame(lamda, columns=['Name', 'Key', 'Value'])
        lamda = lamda.pivot(index='Name', columns='Key', values='Value')
        return lamda.to_numpy()


class DDF(DEA):
    def __init__(self,  y, x, b=None, gy=[1], gx=[1], gb=None, rts=RTS_VRS, yref=None, xref=None, bref=None):
        """DEA: Directional distance function

        Args:
            y (float): output variable. 
            x (float): input variables.
            b (float), optional): undesirable output variables. Defaults to None.
            gy (list, optional): output directional vector. Defaults to [1].
            gx (list, optional): input directional vector. Defaults to [1].
            gb (list, optional): undesirable output directional vector. Defaults to None.
            rts (String): RTS_VRS (variable returns to scale) or RTS_CRS (constant returns to scale)
            yref (String, optional): reference output. Defaults to None.
            xref (String, optional): reference inputs. Defaults to None.
            bref (String, optional): reference undesirable output. Defaults to None.
        """
        # Initialize DEA model
        self.__model__ = ConcreteModel()

        self.y, self.x, self.b, self.gy, self.gx, self.gb = tools.assert_valid_direciontal_data(
            y, x, b, gy, gx, gb)
        self.rts = rts

        if type(yref) != type(None):
            self.yref, self.xref, self.bref = tools.assert_valid_reference_data_with_bad_outputs(
                self.y, self.x, self.b, yref, xref, bref)
        else:
            self.yref, self.xref, self.bref = self.y, self.x, self.b
        self.__model__.R = Set(initialize=range(len(self.yref)))

        # Initialize sets
        self.__model__.I = Set(initialize=range(len(self.y)))
        self.__model__.J = Set(initialize=range(len(self.x[0])))
        self.__model__.K = Set(initialize=range(len(self.y[0])))
        if type(b) != type(None):
            self.__model__.L = Set(initialize=range(len(self.b[0])))

        # Initialize variable
        self.__model__.theta = Var(
            self.__model__.I, doc='directional distance')

        self.__model__.lamda = Var(self.__model__.I, self.__model__.R, bounds=(
            0.0, None), doc='intensity variables')

        # Setup the objective function and constraints
        self.__model__.objective = Objective(
            rule=self._DEA__objective_rule(), sense=maximize, doc='objective function')
        self.__model__.input = Constraint(
            self.__model__.I, self.__model__.J, rule=self.__input_rule(), doc='input constraint')
        self.__model__.output = Constraint(
            self.__model__.I, self.__model__.K, rule=self.__output_rule(), doc='output constraint')

        if type(b) != type(None):
            self.__model__.undesirable_output = Constraint(
                self.__model__.I, self.__model__.L, rule=self.__undesirable_output_rule(), doc='undesirable output constraint')

        if self.rts == RTS_VRS:
            self.__model__.vrs = Constraint(
                self.__model__.I, rule=self.__vrs_rule(), doc='various return to scale rule')

        # Optimize model
        self.optimization_status, self.problem_status = 0, 0

    def __input_rule(self):
        """Return the proper input constraint"""
        def input_rule(model, o, j):
            return self.x[o][j] - model.theta[o]*self.gx[j] >= sum(model.lamda[o, r] * self.xref[r][j] for r in model.R)
        return input_rule

    def __output_rule(self):
        """Return the proper output constraint"""
        def output_rule(model, o, k):
            return self.y[o][k] + model.theta[o]*self.gy[k] <= sum(model.lamda[o, r] * self.yref[r][k] for r in model.R)
        return output_rule

    def __undesirable_output_rule(self):
        """Return the proper undesirable output constraint"""
        def undesirable_output_rule(model, o, l):
            return self.b[o][l] - model.theta[o]*self.gb[l] == sum(model.lamda[o, r] * self.bref[r][l] for r in model.R)
        return undesirable_output_rule

    def __vrs_rule(self):
        """Return the VRS constraint"""
        def vrs_rule(model, o):
            return sum(model.lamda[o, r] for r in model.R) == 1
        return vrs_rule


class DUAL(DEA):
    def __init__(self, y, x, orient, rts, yref=None, xref=None):
        """DEA: Multiplier problem

        Args:
            y (float): output variable. 
            x (float): input variables.
            orient (String): ORIENT_IO (input orientation) or ORIENT_OO (output orientation)
            rts (String): RTS_VRS (variable returns to scale) or RTS_CRS (constant returns to scale)
            yref (String, optional): reference output. Defaults to None.
            xref (String, optional): reference inputs. Defaults to None.
        """
        # Initialize DEA model
        self.__model__ = ConcreteModel()

        self.y, self.x = tools.assert_valid_mupltiple_y_data(y, x)
        self.orient, self.rts = orient, rts

        if type(yref) != type(None):
            self.yref, self.xref = tools.assert_valid_reference_data(
                self.y, self.x, yref, xref)
        else:
            self.yref, self.xref = self.y, self.x
        self.__model__.R = Set(initialize=range(len(self.yref)))

        # Initialize sets
        self.__model__.I = Set(initialize=range(len(self.y)))
        self.__model__.J = Set(initialize=range(len(self.x[0])))
        self.__model__.K = Set(initialize=range(len(self.y[0])))

        # Initialize variable
        self.__model__.nu = Var(self.__model__.I, self.__model__.J, bounds=(
            0.0, None), doc='multiplier x')
        self.__model__.mu = Var(self.__model__.I, self.__model__.K, bounds=(
            0.0, None), doc='multiplier y')
        if self.rts == RTS_VRS:
            self.__model__.omega = Var(
                self.__model__.I, doc='variable return to scale')

        # Setup the objective function and constraints
        if self.orient == ORIENT_IO:
            self.__model__.objective = Objective(
                rule=self.__objective_rule(), sense=maximize, doc='objective function')
        else:
            self.__model__.objective = Objective(
                rule=self.__objective_rule(), sense=minimize, doc='objective function')
        self.__model__.first = Constraint(
            self.__model__.I, self.__model__.R, rule=self.__first_rule(), doc='technology constraint')
        self.__model__.second = Constraint(
            self.__model__.I, rule=self.__second_rule(), doc='normalization constraint')

        # Optimize model
        self.optimization_status, self.problem_status = 0, 0

    def __objective_rule(self):
        """Return the proper objective function"""
        if self.orient == ORIENT_IO:
            def objective_rule(model):
                if self.rts == RTS_VRS:
                    return sum(sum(model.mu[o, k] * self.y[o][k] for o in model.I) for k in model.K) + sum(model.omega[o] for o in model.I)
                elif self.rts == RTS_CRS:
                    return sum(sum(model.mu[o, k] * self.y[o][k] for o in model.I) for k in model.K)
            return objective_rule
        elif self.orient == ORIENT_OO:
            def objective_rule(model):
                if self.rts == RTS_VRS:
                    return sum(sum(model.nu[o, j] * self.x[o][j] for o in model.I) for j in model.J) + sum(model.omega[o] for o in model.I)
                elif self.rts == RTS_CRS:
                    return sum(sum(model.nu[o, j] * self.x[o][j] for o in model.I) for j in model.J)
            return objective_rule

    def __first_rule(self):
        """Return the proper technology constraint"""
        if self.orient == ORIENT_IO:
            if self.rts == RTS_VRS:
                def first_rule(model, o, r):
                    return sum(model.mu[o, k] * self.yref[r][k] for k in model.K) - sum(model.nu[o, j] * self.xref[r][j] for j in model.J) + model.omega[o] <= 0
                return first_rule
            elif self.rts == RTS_CRS:
                def first_rule(model, o, r):
                    return sum(model.mu[o, k] * self.yref[r][k] for k in model.K) - sum(model.nu[o, j] * self.xref[r][j] for j in model.J) <= 0
                return first_rule
        elif self.orient == ORIENT_OO:
            if self.rts == RTS_VRS:
                def first_rule(model, o, r):
                    return sum(model.nu[o, j] * self.xref[r][j] for j in model.J) - sum(model.mu[o, k] * self.yref[r][k] for k in model.K) + model.omega[o] >= 0
                return first_rule
            elif self.rts == RTS_CRS:
                def first_rule(model, o, r):
                    return sum(model.nu[o, j] * self.xref[r][j] for j in model.J) - sum(model.mu[o, k] * self.yref[r][k] for k in model.K) >= 0
                return first_rule

    def __second_rule(self):
        """Return the proper normalization constraint"""
        if self.orient == ORIENT_IO:
            def second_rule(model, o):
                return sum(model.nu[o, j] * self.x[o][j] for j in model.J) == 1
            return second_rule
        elif self.orient == ORIENT_OO:
            def second_rule(model, o):
                return sum(model.mu[o, k] * self.y[o][k] for k in model.K) == 1
            return second_rule

    def display_mu(self):
        """Display mu value"""
        tools.assert_optimized(self.optimization_status)
        self.__model__.mu.display()

    def display_nu(self):
        """Display nu value"""
        tools.assert_optimized(self.optimization_status)
        self.__model__.nu.display()

    def display_omega(self):
        """Display omega value"""
        tools.assert_optimized(self.optimization_status)
        tools.assert_various_return_to_scale_omega(self.rts)
        self.__model__.omega.display()

    def get_mu(self):
        """Return mu value by array"""
        tools.assert_optimized(self.optimization_status)
        mu = np.asarray([i + tuple([j]) for i, j in zip(list(self.__model__.mu),
                                                        list(self.__model__.mu[:, :].value))])
        mu = pd.DataFrame(mu, columns=['Name', 'Key', 'Value'])
        mu = mu.pivot(index='Name', columns='Key', values='Value')
        return mu.to_numpy()

    def get_nu(self):
        """Return nu value by array"""
        tools.assert_optimized(self.optimization_status)
        nu = np.asarray([i + tuple([j]) for i, j in zip(list(self.__model__.nu),
                                                        list(self.__model__.nu[:, :].value))])
        nu = pd.DataFrame(nu, columns=['Name', 'Key', 'Value'])
        nu = nu.pivot(index='Name', columns='Key', values='Value')
        return nu.to_numpy()

    def get_omega(self):
        """Return omega value by array"""
        tools.assert_optimized(self.optimization_status)
        tools.assert_various_return_to_scale_omega(self.rts)
        omega = list(self.__model__.omega[:].value)
        return np.asarray(omega)

    def get_efficiency(self):
        """Return efficiency value by array"""
        tools.assert_optimized(self.optimization_status)
        if self.orient == ORIENT_IO:
            if self.rts == RTS_CRS:
                return (np.sum(self.get_mu()*self.y, axis=1)).reshape(len(self.y), 1)
            elif self.rts == RTS_VRS:
                return (np.sum(self.get_mu()*self.y, axis=1)).reshape(len(self.y), 1) + self.get_omega().reshape(len(self.y), 1)
        elif self.orient == ORIENT_OO:
            if self.rts == RTS_CRS:
                return (np.sum(self.get_nu()*self.x, axis=1)).reshape(len(self.x), 1)
            elif self.rts == RTS_VRS:
                return (np.sum(self.get_nu()*self.x, axis=1)).reshape(len(self.x), 1) + self.get_omega().reshape(len(self.x), 1)



# import dependencies
import numpy as np
import pandas as pd
from .utils import CQERG1, CQERG2, CQERZG1, CQERZG2, sweet, tools, interpolation
from .constant import CET_ADDI, CET_MULT, FUN_PROD, FUN_COST, RTS_CRS, RTS_VRS, OPT_LOCAL, OPT_DEFAULT
import time


class CQRG:
    """Convex quantile regression (CQR) with Genetic algorithm
    """

    def __init__(self, y, x, tau, z=None, cet=CET_ADDI, fun=FUN_PROD, rts=RTS_VRS):
        """CQRG model

        Args:
            y (float): output variable. 
            x (float): input variables.
            tau (float): quantile.
            z (float, optional): Contextual variable(s). Defaults to None.
            cet (String, optional): CET_ADDI (additive composite error term) or CET_MULT (multiplicative composite error term). Defaults to CET_ADDI.
            fun (String, optional): FUN_PROD (production frontier) or FUN_COST (cost frontier). Defaults to FUN_PROD.
            rts (String, optional): RTS_VRS (variable returns to scale) or RTS_CRS (constant returns to scale). Defaults to RTS_VRS.
        """
        # TODO(error/warning handling): Check the configuration of the model exist
        self.cutactive = sweet.sweet(x)
        self.y, self.x, self.z = tools.assert_valid_basic_data(y, x, z)
        self.tau, self.cet, self.fun, self.rts = tau, cet, fun, rts

        # active (added) violated concavity constraint by iterative procedure
        self.active = np.zeros((len(x), len(x)))
        # violated concavity constraint
        self.active2 = np.zeros((len(x), len(x)))

        # Optimize model
        self.optimization_status, self.problem_status = 0, 0

    def optimize(self, email=OPT_LOCAL, solver=OPT_DEFAULT):
        """Optimize the function by requested method"""
        # TODO(error/warning handling): Check problem status after optimization
        self.t0 = time.time()
        if type(self.z) != type(None):
            model1 = CQERZG1.CQRZG1(
                self.y, self.x, self.z, self.tau, self.cutactive, self.cet, self.fun, self.rts)
        else:
            model1 = CQERG1.CQRG1(
                self.y, self.x, self.tau, self.cutactive, self.cet, self.fun, self.rts)
        model1.optimize(email, solver)
        self.alpha = model1.get_alpha()
        self.beta = model1.get_beta()
        self.__model__ = model1.__model__

        self.count = 0
        while self.__convergence_test(self.alpha, self.beta) > 0.0001:
            if type(self.z) != type(None):
                model2 = CQERZG2.CQRZG2(
                    self.y, self.x, self.z, self.tau, self.active, self.cutactive, self.cet, self.fun, self.rts)
            else:
                model2 = CQERG2.CQRG2(
                    self.y, self.x, self.tau, self.active, self.cutactive, self.cet, self.fun, self.rts)
            model2.optimize(email, solver)
            self.alpha = model2.get_alpha()
            self.beta = model2.get_beta()
            # TODO: Replace print with log system
            print("Genetic Algorithm Convergence : %8f" %
                  (self.__convergence_test(self.alpha, self.beta)))
            self.__model__ = model2.__model__
            self.count += 1
        self.optimization_status = 1
        self.tt = time.time() - self.t0

    def __convergence_test(self, alpha, beta):
        x = np.asarray(self.x)
        activetmp1 = 0.0

        # go into the loop
        for i in range(len(x)):
            activetmp = 0.0
            # go into the sub-loop and find the violated concavity constraints
            for j in range(len(x)):
                if self.cet == CET_ADDI:
                    if self.rts == RTS_VRS:
                        if self.fun == FUN_PROD:
                            self.active2[i, j] = alpha[i] + np.sum(beta[i, :] * x[i, :]) - \
                                alpha[j] - np.sum(beta[j, :] * x[i, :])
                        elif self.fun == FUN_COST:
                            self.active2[i, j] = - alpha[i] - np.sum(beta[i, :] * x[i, :]) + \
                                alpha[j] + np.sum(beta[j, :] * x[i, :])
                    if self.rts == RTS_CRS:
                        if self.fun == FUN_PROD:
                            self.active2[i, j] = np.sum(beta[i, :] * x[i, :]) \
                                - np.sum(beta[j, :] * x[i, :])
                        elif self.fun == FUN_COST:
                            self.active2[i, j] = - np.sum(beta[i, :] * x[i, :]) \
                                + np.sum(beta[j, :] * x[i, :])
                if self.cet == CET_MULT:
                    if self.rts == RTS_VRS:
                        if self.fun == FUN_PROD:
                            self.active2[i, j] = alpha[i] + np.sum(beta[i, :] * x[i, :]) - \
                                alpha[j] - np.sum(beta[j, :] * x[i, :])
                        elif self.fun == FUN_COST:
                            self.active2[i, j] = - alpha[i] - np.sum(beta[i, :] * x[i, :]) + \
                                alpha[j] + np.sum(beta[j, :] * x[i, :])
                    if self.rts == RTS_CRS:
                        if self.fun == FUN_PROD:
                            self.active2[i, j] = np.sum(beta[i, :] * x[i, :]) - \
                                np.sum(beta[j, :] * x[i, :])
                        elif self.fun == FUN_COST:
                            self.active2[i, j] = - np.sum(beta[i, :] * x[i, :]) + \
                                np.sum(beta[j, :] * x[i, :])
                if self.active2[i, j] > activetmp:
                    activetmp = self.active2[i, j]
            # find the maximal violated constraint in sub-loop and added into the active matrix
            for j in range(len(x)):
                if self.active2[i, j] >= activetmp and activetmp > 0:
                    self.active[i, j] = 1
            if activetmp > activetmp1:
                activetmp1 = activetmp
        return activetmp

    def display_status(self):
        """Display the status of problem"""
        print(self.optimization_status)

    def display_alpha(self):
        """Display alpha value"""
        tools.assert_optimized(self.optimization_status)
        tools.assert_various_return_to_scale(self.rts)
        self.__model__.alpha.display()

    def display_beta(self):
        """Display beta value"""
        tools.assert_optimized(self.optimization_status)
        self.__model__.beta.display()

    def display_lamda(self):
        """Display lamda value"""
        tools.assert_optimized(self.optimization_status)
        tools.assert_contextual_variable(self.z)
        self.__model__.lamda.display()

    def display_residual(self):
        """Dispaly residual value"""
        tools.assert_optimized(self.optimization_status)
        self.__model__.epsilon.display()

    def display_positive_residual(self):
        """Dispaly positive residual value"""
        tools.assert_optimized(self.optimization_status)
        self.__model__.epsilon_plus.display()

    def display_negative_residual(self):
        """Dispaly negative residual value"""
        tools.assert_optimized(self.optimization_status)
        self.__model__.epsilon_minus.display()

    def get_status(self):
        """Return status"""
        return self.optimization_status

    def get_alpha(self):
        """Return alpha value by array"""
        tools.assert_optimized(self.optimization_status)
        tools.assert_various_return_to_scale(self.rts)
        alpha = list(self.__model__.alpha[:].value)
        return np.asarray(alpha)

    def get_beta(self):
        """Return beta value by array"""
        tools.assert_optimized(self.optimization_status)
        beta = np.asarray([i + tuple([j]) for i, j in zip(list(self.__model__.beta),
                                                          list(self.__model__.beta[:, :].value))])
        beta = pd.DataFrame(beta, columns=['Name', 'Key', 'Value'])
        beta = beta.pivot(index='Name', columns='Key', values='Value')
        return beta.to_numpy()

    def get_residual(self):
        """Return residual value by array"""
        tools.assert_optimized(self.optimization_status)
        residual = list(self.__model__.epsilon[:].value)
        return np.asarray(residual)

    def get_positive_residual(self):
        """Return positive residual value by array"""
        tools.assert_optimized(self.optimization_status)
        residual_plus = list(self.__model__.epsilon_plus[:].value)
        return np.asarray(residual_plus)

    def get_negative_residual(self):
        """Return negative residual value by array"""
        tools.assert_optimized(self.optimization_status)
        residual_minus = list(self.__model__.epsilon_minus[:].value)
        return np.asarray(residual_minus)

    def get_lamda(self):
        """Return beta value by array"""
        tools.assert_optimized(self.optimization_status)
        tools.assert_contextual_variable(self.z)
        lamda = list(self.__model__.lamda[:].value)
        return np.asarray(lamda)

    def get_frontier(self):
        """Return estimated frontier value by array"""
        tools.assert_optimized(self.optimization_status)
        if self.cet == CET_MULT and type(self.z) == type(None):
            frontier = np.asarray(list(self.__model__.frontier[:].value))+1
        elif self.cet == CET_MULT and type(self.z) != type(None):
            frontier = list(np.divide(self.y, np.exp(
                self.get_residual()+self.get_lamda()*np.asarray(self.z)[:, 0])) - 1)
        elif self.cet == CET_ADDI:
            frontier = np.asarray(self.y) - self.get_residual()
        return np.asarray(frontier)

    def get_totalconstr(self):
        """Return the number of total constraints"""
        tools.assert_optimized(self.optimization_status)
        activeconstr = np.sum(self.active) - np.trace(self.active)
        cutactiveconstr = np.sum(self.cutactive) - np.trace(self.cutactive)
        totalconstr = activeconstr + cutactiveconstr + 2 * len(self.active) + 1
        return totalconstr

    def get_runningtime(self):
        """Return the running time"""
        tools.assert_optimized(self.optimization_status)
        return self.tt

    def get_blocks(self):
        """Return the number of blocks"""
        tools.assert_optimized(self.optimization_status)
        return self.count

    def get_predict(self, x_test):
        """Return the estimated function in testing sample"""
        tools.assert_optimized(self.optimization_status)
        if self.rts == RTS_VRS:
            alpha,  beta = self.get_alpha(), self.get_beta()
        elif self.rts == RTS_CRS:
            alpha, beta = np.zeros((self.get_beta()).shape[0]), self.get_beta()
        return interpolation.interpolation(alpha, beta, x_test, fun=self.fun)


class CERG:
    """Convex expectile regression (CER) with Genetic algorithm
    """

    def __init__(self, y, x, tau, z=None, cet=CET_ADDI, fun=FUN_PROD, rts=RTS_VRS):
        """CERG model

        Args:
            y (float): output variable. 
            x (float): input variables.
            tau (float): quantile.
            z (float, optional): Contextual variable(s). Defaults to None.
            cet (String, optional): CET_ADDI (additive composite error term) or CET_MULT (multiplicative composite error term). Defaults to CET_ADDI.
            fun (String, optional): FUN_PROD (production frontier) or FUN_COST (cost frontier). Defaults to FUN_PROD.
            rts (String, optional): RTS_VRS (variable returns to scale) or RTS_CRS (constant returns to scale). Defaults to RTS_VRS.
        """
        # TODO(error/warning handling): Check the configuration of the model exist
        self.cutactive = sweet.sweet(x)
        self.y, self.x, self.z = tools.assert_valid_basic_data(y, x, z)
        self.tau, self.cet, self.fun, self.rts = tau, cet, fun, rts

        # active (added) violated concavity constraint by iterative procedure
        self.active = np.zeros((len(x), len(x)))
        # violated concavity constraint
        self.active2 = np.zeros((len(x), len(x)))

        # Optimize model
        self.optimization_status, self.problem_status = 0, 0

    def optimize(self, email=OPT_LOCAL, solver=OPT_DEFAULT):
        """Optimize the function by requested method"""
        # TODO(error/warning handling): Check problem status after optimization
        self.t0 = time.time()
        if type(self.z) != type(None):
            model1 = CQERZG1.CERZG1(
                self.y, self.x, self.z, self.tau, self.cutactive, self.cet, self.fun, self.rts)
        else:
            model1 = CQERG1.CERG1(
                self.y, self.x, self.tau, self.cutactive, self.cet, self.fun, self.rts)
        model1.optimize(email, solver)
        self.alpha = model1.get_alpha()
        self.beta = model1.get_beta()
        self.__model__ = model1.__model__

        self.count = 0
        while self.__convergence_test(self.alpha, self.beta) > 0.0001:
            if type(self.z) != type(None):
                model2 = CQERZG2.CERZG2(
                    self.y, self.x, self.z, self.tau, self.active, self.cutactive, self.cet, self.fun, self.rts)
            else:
                model2 = CQERG2.CERG2(
                    self.y, self.x, self.tau, self.active, self.cutactive, self.cet, self.fun, self.rts)
            model2.optimize(email, solver)
            self.alpha = model2.get_alpha()
            self.beta = model2.get_beta()
            # TODO: Replace print with log system
            print("Genetic Algorithm Convergence : %8f" %
                  (self.__convergence_test(self.alpha, self.beta)))
            self.__model__ = model2.__model__
            self.count += 1
        self.optimization_status = 1
        self.tt = time.time() - self.t0

    def __convergence_test(self, alpha, beta):
        x = np.asarray(self.x)
        activetmp1 = 0.0

        # go into the loop
        for i in range(len(x)):
            activetmp = 0.0
            # go into the sub-loop and find the violated concavity constraints
            for j in range(len(x)):
                if self.cet == CET_ADDI:
                    if self.rts == RTS_VRS:
                        if self.fun == FUN_PROD:
                            self.active2[i, j] = alpha[i] + np.sum(beta[i, :] * x[i, :]) - \
                                alpha[j] - np.sum(beta[j, :] * x[i, :])
                        elif self.fun == FUN_COST:
                            self.active2[i, j] = - alpha[i] - np.sum(beta[i, :] * x[i, :]) + \
                                alpha[j] + np.sum(beta[j, :] * x[i, :])
                    if self.rts == RTS_CRS:
                        if self.fun == FUN_PROD:
                            self.active2[i, j] = np.sum(beta[i, :] * x[i, :]) \
                                - np.sum(beta[j, :] * x[i, :])
                        elif self.fun == FUN_COST:
                            self.active2[i, j] = - np.sum(beta[i, :] * x[i, :]) \
                                + np.sum(beta[j, :] * x[i, :])
                if self.cet == CET_MULT:
                    if self.rts == RTS_VRS:
                        if self.fun == FUN_PROD:
                            self.active2[i, j] = alpha[i] + np.sum(beta[i, :] * x[i, :]) - \
                                alpha[j] - np.sum(beta[j, :] * x[i, :])
                        elif self.fun == FUN_COST:
                            self.active2[i, j] = - alpha[i] - np.sum(beta[i, :] * x[i, :]) + \
                                alpha[j] + np.sum(beta[j, :] * x[i, :])
                    if self.rts == RTS_CRS:
                        if self.fun == FUN_PROD:
                            self.active2[i, j] = np.sum(beta[i, :] * x[i, :]) - \
                                np.sum(beta[j, :] * x[i, :])
                        elif self.fun == FUN_COST:
                            self.active2[i, j] = - np.sum(beta[i, :] * x[i, :]) + \
                                np.sum(beta[j, :] * x[i, :])
                if self.active2[i, j] > activetmp:
                    activetmp = self.active2[i, j]
            # find the maximal violated constraint in sub-loop and added into the active matrix
            for j in range(len(x)):
                if self.active2[i, j] >= activetmp and activetmp > 0:
                    self.active[i, j] = 1
            if activetmp > activetmp1:
                activetmp1 = activetmp
        return activetmp

    def display_status(self):
        """Display the status of problem"""
        print(self.optimization_status)

    def display_alpha(self):
        """Display alpha value"""
        tools.assert_optimized(self.optimization_status)
        tools.assert_various_return_to_scale(self.rts)
        self.__model__.alpha.display()

    def display_beta(self):
        """Display beta value"""
        tools.assert_optimized(self.optimization_status)
        self.__model__.beta.display()

    def display_lamda(self):
        """Display lamda value"""
        tools.assert_optimized(self.optimization_status)
        tools.assert_contextual_variable(self.z)
        self.__model__.lamda.display()

    def display_residual(self):
        """Dispaly residual value"""
        tools.assert_optimized(self.optimization_status)
        self.__model__.epsilon.display()

    def display_positive_residual(self):
        """Dispaly positive residual value"""
        tools.assert_optimized(self.optimization_status)
        self.__model__.epsilon_plus.display()

    def display_negative_residual(self):
        """Dispaly negative residual value"""
        tools.assert_optimized(self.optimization_status)
        self.__model__.epsilon_minus.display()

    def get_status(self):
        """Return status"""
        return self.optimization_status

    def get_alpha(self):
        """Return alpha value by array"""
        tools.assert_optimized(self.optimization_status)
        tools.assert_various_return_to_scale(self.rts)
        alpha = list(self.__model__.alpha[:].value)
        return np.asarray(alpha)

    def get_beta(self):
        """Return beta value by array"""
        tools.assert_optimized(self.optimization_status)
        beta = np.asarray([i + tuple([j]) for i, j in zip(list(self.__model__.beta),
                                                          list(self.__model__.beta[:, :].value))])
        beta = pd.DataFrame(beta, columns=['Name', 'Key', 'Value'])
        beta = beta.pivot(index='Name', columns='Key', values='Value')
        return beta.to_numpy()

    def get_residual(self):
        """Return residual value by array"""
        tools.assert_optimized(self.optimization_status)
        residual = list(self.__model__.epsilon[:].value)
        return np.asarray(residual)

    def get_positive_residual(self):
        """Return positive residual value by array"""
        tools.assert_optimized(self.optimization_status)
        residual_plus = list(self.__model__.epsilon_plus[:].value)
        return np.asarray(residual_plus)

    def get_negative_residual(self):
        """Return negative residual value by array"""
        tools.assert_optimized(self.optimization_status)
        residual_minus = list(self.__model__.epsilon_minus[:].value)
        return np.asarray(residual_minus)

    def get_lamda(self):
        """Return beta value by array"""
        tools.assert_optimized(self.optimization_status)
        tools.assert_contextual_variable(self.z)
        lamda = list(self.__model__.lamda[:].value)
        return np.asarray(lamda)

    def get_frontier(self):
        """Return estimated frontier value by array"""
        tools.assert_optimized(self.optimization_status)
        if self.cet == CET_MULT and type(self.z) == type(None):
            frontier = np.asarray(list(self.__model__.frontier[:].value))+1
        elif self.cet == CET_MULT and type(self.z) != type(None):
            frontier = list(np.divide(self.y, np.exp(
                self.get_residual()+self.get_lamda()*np.asarray(self.z)[:, 0])) - 1)
        elif self.cet == CET_ADDI:
            frontier = np.asarray(self.y) - self.get_residual()
        return np.asarray(frontier)

    def get_totalconstr(self):
        """Return the number of total constraints"""
        tools.assert_optimized(self.optimization_status)
        activeconstr = np.sum(self.active) - np.trace(self.active)
        cutactiveconstr = np.sum(self.cutactive) - np.trace(self.cutactive)
        totalconstr = activeconstr + cutactiveconstr + 2 * len(self.active) + 1
        return totalconstr

    def get_runningtime(self):
        """Return the running time"""
        tools.assert_optimized(self.optimization_status)
        return self.tt

    def get_blocks(self):
        """Return the number of blocks"""
        tools.assert_optimized(self.optimization_status)
        return self.count

    def get_predict(self, x_test):
        """Return the estimated function in testing sample"""
        tools.assert_optimized(self.optimization_status)
        if self.rts == RTS_VRS:
            alpha,  beta = self.get_alpha(), self.get_beta()
        elif self.rts == RTS_CRS:
            alpha, beta = np.zeros((self.get_beta()).shape[0]), self.get_beta()
        return interpolation.interpolation(alpha, beta, x_test, fun=self.fun)



# import dependencies
from pyomo.environ import ConcreteModel, Set, Var, Objective, minimize, Constraint
from pyomo.core.expr.numvalue import NumericValue
import numpy as np
import pandas as pd

from .constant import CET_ADDI, FUN_PROD, FUN_COST, OPT_DEFAULT, OPT_LOCAL
from .utils import tools, interpolation


class CSVR:
    """Convex Support Vector Regression (CSVR)
    """

    def __init__(self, y, x, fun=FUN_PROD, epsilon=0.01, C=2):
        """CSVR model

        Args:
            y (float): output variable. 
            x (float): input variables.
            epsilon (float): epsilon-loss function.
            C (float): Regularization parameter.
            fun (String, optional): FUN_PROD (production frontier) or FUN_COST (cost frontier). Defaults to FUN_PROD.
        """
        # TODO(error/warning handling): Check the configuration of the model exist
        self.y, self.x, self.z = tools.assert_valid_basic_data(y, x, z=None)
        self.epsilon, self.C, self.cet, self.fun = epsilon, C, CET_ADDI, fun

        # Initialize the CNLS model
        self.__model__ = ConcreteModel()

        # Initialize the sets
        self.__model__.I = Set(initialize=range(len(self.y)))
        self.__model__.J = Set(initialize=range(len(self.x[0])))

        # Initialize the variables
        self.__model__.alpha = Var(self.__model__.I, doc='alpha')
        self.__model__.beta = Var(self.__model__.I,
                                  self.__model__.J,
                                  doc='beta')
        self.__model__.ksia = Var(self.__model__.I,
                                  bounds=(0.0, None),
                                  doc='Ksi a')
        self.__model__.ksib = Var(self.__model__.I,
                                  bounds=(0.0, None),
                                  doc='Ksi b')

        # Setup the objective function and constraints
        self.__model__.objective = Objective(rule=self.__objective_rule(),
                                             sense=minimize,
                                             doc='objective function')

        self.__model__.regression_rule1 = Constraint(self.__model__.I,
                                                     rule=self.__regression_rule1(),
                                                     doc='regression equation')

        self.__model__.regression_rule2 = Constraint(self.__model__.I,
                                                     rule=self.__regression_rule2(),
                                                     doc='regression equation')

        self.__model__.afriat_rule = Constraint(self.__model__.I,
                                                self.__model__.I,
                                                rule=self.__afriat_rule(),
                                                doc='afriat inequality')

        # Optimize model
        self.optimization_status, self.problem_status = 0, 0

    def optimize(self, email=OPT_LOCAL, solver=OPT_DEFAULT):
        """Optimize the function by requested method

        Args:
            email (string): The email address for remote optimization. It will optimize locally if OPT_LOCAL is given.
            solver (string): The solver chosen for optimization. It will optimize with default solver if OPT_DEFAULT is given.
        """
        # TODO(error/warning handling): Check problem status after optimization
        self.problem_status, self.optimization_status = tools.optimize_model(
            self.__model__, email, self.cet, solver)

    def __objective_rule(self):
        """Return the proper objective function"""

        def objective_rule(model):
            return self.C * (sum(model.ksia[i] for i in model.I) + sum(model.ksib[i] for i in model.I)) + \
                sum(model.beta[i, j]**2 for i in model.I for j in model.J)

        return objective_rule

    def __regression_rule1(self):
        """Return the proper regression constraint"""

        def regression_rule1(model, i):
            return self.y[i] - sum(model.beta[i, j] * self.x[i][j] for j in model.J) - model.alpha[i]\
                <= self.epsilon + model.ksia[i]

        return regression_rule1

    def __regression_rule2(self):
        """Return the proper regression constraint"""

        def regression_rule2(model, i):
            return sum(model.beta[i, j] * self.x[i][j] for j in model.J) + model.alpha[i] - self.y[i]\
                <= self.epsilon + model.ksib[i]

        return regression_rule2

    def __afriat_rule(self):
        """Return the proper afriat inequality constraint"""
        if self.fun == FUN_PROD:
            __operator = NumericValue.__le__
        elif self.fun == FUN_COST:
            __operator = NumericValue.__ge__

        def afriat_rule(model, i, h):
            if i == h:
                return Constraint.Skip
            return __operator(
                model.alpha[i] + sum(model.beta[i, j] * self.x[i][j]
                                     for j in model.J),
                model.alpha[h] + sum(model.beta[h, j] * self.x[i][j]
                                     for j in model.J))

        return afriat_rule

    def display_status(self):
        """Display the status of problem"""
        tools.assert_optimized(self.optimization_status)
        print(self.display_status)

    def display_alpha(self):
        """Display alpha value"""
        tools.assert_optimized(self.optimization_status)
        self.__model__.alpha.display()

    def display_beta(self):
        """Display beta value"""
        tools.assert_optimized(self.optimization_status)
        self.__model__.beta.display()

    def get_status(self):
        """Return status"""
        return self.optimization_status

    def get_alpha(self):
        """Return alpha value by array"""
        tools.assert_optimized(self.optimization_status)
        alpha = list(self.__model__.alpha[:].value)
        return np.asarray(alpha)

    def get_beta(self):
        """Return beta value by array"""
        tools.assert_optimized(self.optimization_status)
        beta = np.asarray([i + tuple([j]) for i, j in zip(list(self.__model__.beta),
                                                          list(self.__model__.beta[:, :].value))])
        beta = pd.DataFrame(beta, columns=['Name', 'Key', 'Value'])
        beta = beta.pivot(index='Name', columns='Key', values='Value')
        return beta.to_numpy()

    def get_predict(self, x_test):
        """Return the estimated function in testing sample"""
        tools.assert_optimized(self.optimization_status)
        alpha, beta = self.get_alpha(), self.get_beta()
        return interpolation.interpolation(alpha, beta, x_test, fun=self.fun)



# import dependencies
from pyomo.environ import ConcreteModel, Set, Var, Objective, minimize, Constraint
from pyomo.core.expr.numvalue import NumericValue
from .constant import FUN_PROD, FUN_COST, RTS_VRS
from . import CNLSDDF, CQER
from .utils import tools


class CQRDDF(CNLSDDF.CNLSDDF, CQER.CQR):
    """Convex quantile regression with directional distance function
    """

    def __init__(self, y, x, b=None, gy=[1], gx=[1], gb=None, fun=FUN_PROD, tau=0.5):
        """CQR DDF 

        Args:
            y (float): output variable.
            x (float): input variables.
            b (float), optional): undesirable output variables. Defaults to None.
            gy (list, optional): output directional vector. Defaults to [1].
            gx (list, optional): input directional vector. Defaults to [1].
            gb (list, optional): undesirable output directional vector. Defaults to None.
            fun (String, optional): FUN_PROD (production frontier) or FUN_COST (cost frontier). Defaults to FUN_PROD.
            tau (float, optional): quantile. Defaults to 0.5.
        """
        # TODO(error/warning handling): Check the configuration of the model exist
        self.y, self.x, self.b, self.gy, self.gx, self.gb = tools.assert_valid_direciontal_data(
            y, x, b, gy, gx, gb)
        self.tau, self.fun, self.rts = tau, fun, RTS_VRS

        # Initialize the CQRDDF model
        self.__model__ = ConcreteModel()

        # Initialize the sets
        self.__model__.I = Set(initialize=range(len(self.y)))
        self.__model__.J = Set(initialize=range(len(self.x[0])))
        self.__model__.Q = Set(initialize=range(len(self.y[0])))

        # Initialize the variables
        self.__model__.alpha = Var(self.__model__.I, doc='alpha')
        self.__model__.beta = Var(
            self.__model__.I, self.__model__.J, bounds=(0.0, None), doc='beta')
        self.__model__.gamma = Var(
            self.__model__.I, self.__model__.Q, bounds=(0.0, None), doc='gamma')

        self.__model__.epsilon = Var(self.__model__.I, doc='residuals')
        self.__model__.epsilon_plus = Var(
            self.__model__.I, bounds=(0.0, None), doc='positive error term')
        self.__model__.epsilon_minus = Var(
            self.__model__.I, bounds=(0.0, None), doc='negative error term')

        if type(self.b) != type(None):
            self.__model__.L = Set(initialize=range(len(self.b[0])))
            self.__model__.delta = Var(
                self.__model__.I, self.__model__.L, bounds=(0.0, None), doc='delta')

        self.__model__.objective = Objective(rule=self._CQR__objective_rule(),
                                             sense=minimize,
                                             doc='objective function')

        self.__model__.error_decomposition = Constraint(self.__model__.I,
                                                        rule=self._CQR__error_decomposition(),
                                                        doc='decompose error term')

        self.__model__.regression_rule = Constraint(self.__model__.I,
                                                    rule=self.__regression_rule(),
                                                    doc='regression equation')

        self.__model__.translation_rule = Constraint(self.__model__.I,
                                                     rule=self._CNLSDDF__translation_property(),
                                                     doc='translation property')

        self.__model__.afriat_rule = Constraint(self.__model__.I,
                                                self.__model__.I,
                                                rule=self.__afriat_rule(),
                                                doc='afriat inequality')

        # Optimize model
        self.optimization_status, self.problem_status = 0, 0

    def __regression_rule(self):
        """Return the proper regression constraint"""
        if type(self.b) == type(None):
            def regression_rule(model, i):
                return sum(model.gamma[i, q] * self.y[i][q] for q in model.Q) \
                    == model.alpha[i] \
                    + sum(model.beta[i, j] * self.x[i][j] for j in model.J) \
                    + model.epsilon[i]

            return regression_rule

        def regression_rule(model, i):
            return sum(model.gamma[i, q] * self.y[i][q] for q in model.Q) \
                == model.alpha[i] \
                + sum(model.beta[i, j] * self.x[i][j] for j in model.J) \
                + sum(model.delta[i, l] * self.b[i][l] for l in model.L) \
                + model.epsilon[i]

        return regression_rule

    def __afriat_rule(self):
        """Return the proper afriat inequality constraint"""
        if self.fun == FUN_PROD:
            __operator = NumericValue.__le__
        elif self.fun == FUN_COST:
            __operator = NumericValue.__ge__

        if type(self.b) == type(None):
            def afriat_rule(model, i, h):
                if i == h:
                    return Constraint.Skip
                return __operator(model.alpha[i]
                                  + sum(model.beta[i, j] * self.x[i][j]
                                        for j in model.J)
                                  - sum(model.gamma[i, q] * self.y[i][q]
                                        for q in model.Q),
                                  model.alpha[h]
                                  + sum(model.beta[h, j] * self.x[i][j]
                                        for j in model.J)
                                  - sum(model.gamma[h, q] * self.y[i][q] for q in model.Q))

            return afriat_rule

        def afriat_rule(model, i, h):
            if i == h:
                return Constraint.Skip
            return __operator(model.alpha[i]
                              + sum(model.beta[i, j] * self.x[i][j]
                                    for j in model.J)
                              + sum(model.delta[i, l] * self.b[i][l]
                                    for l in model.L)
                              - sum(model.gamma[i, q] * self.y[i][q]
                                    for q in model.Q),
                              model.alpha[h]
                              + sum(model.beta[h, j] * self.x[i][j]
                                    for j in model.J)
                              + sum(model.delta[h, l] * self.b[i][l]
                                    for l in model.L)
                              - sum(model.gamma[h, q] * self.y[i][q] for q in model.Q))

        return afriat_rule


class CERDDF(CQRDDF):
    """Convex expectile regression with DDF formulation
    """

    def __init__(self, y, x, b=None, gy=[1], gx=[1], gb=None, fun=FUN_PROD, tau=0.5):
        """CER DDF 

        Args:
            y (float): output variable.
            x (float): input variables.
            b (float), optional): undesirable output variables. Defaults to None.
            gy (list, optional): output directional vector. Defaults to [1].
            gx (list, optional): input directional vector. Defaults to [1].
            gb (list, optional): undesirable output directional vector. Defaults to None.
            fun (String, optional): FUN_PROD (production frontier) or FUN_COST (cost frontier). Defaults to FUN_PROD.
            tau (float, optional): expectile. Defaults to 0.5.
        """
        super().__init__(y, x, b, gy, gx, gb, fun, tau)
        self.__model__.objective.deactivate()
        self.__model__.squared_objective = Objective(
            rule=self.__squared_objective_rule(), sense=minimize, doc='squared objective rule')

    def __squared_objective_rule(self):
        def squared_objective_rule(model):
            return self.tau * sum(model.epsilon_plus[i] ** 2 for i in model.I) \
                + (1 - self.tau) * \
                sum(model.epsilon_minus[i] ** 2 for i in model.I)

        return squared_objective_rule



# import dependencies
from pyomo.environ import Constraint, Var
from . import CNLS
from .constant import CET_ADDI, FUN_PROD, RTS_VRS


class pCNLS(CNLS.CNLS):
    """penalized Convex Nonparametric Least Square (pCNLS)
    """

    def __init__(self, y, x, eta, z=None, cet=CET_ADDI, fun=FUN_PROD, rts=RTS_VRS, penalty=1):
        """pCNLS model

        Args:
            y (float): output variable. 
            x (float): input variables.
            eta (float): regularization parameter.
            z (float, optional): Contextual variable(s). Defaults to None.
            cet (String, optional): CET_ADDI (additive composite error term) or CET_MULT (multiplicative composite error term). Defaults to CET_ADDI.
            fun (String, optional): FUN_PROD (production frontier) or FUN_COST (cost frontier). Defaults to FUN_PROD.
            rts (String, optional): RTS_VRS (variable returns to scale) or RTS_CRS (constant returns to scale). Defaults to RTS_VRS.
            penalty (int, optional): penalty=1 (L1 norm), penalty=2 (L2 norm), and penalty=3 (Lipschitz norm). Defaults to 1.
        """
        self.eta = eta
        CNLS.CNLS.__init__(self, y, x, z, cet, fun, rts)
        self.__model__.zeta = Var(
            self.__model__.I, self.__model__.J, doc='zeta')
        if penalty == 1:
            self.__model__.l1_norm1 = Constraint(self.__model__.I,
                                                 self.__model__.J,
                                                 rule=self.__l1_rule1(),
                                                 doc='L1 norm 1')
            self.__model__.l1_norm2 = Constraint(self.__model__.I,
                                                 self.__model__.J,
                                                 rule=self.__l1_rule2(),
                                                 doc='L1 norm 2')
            self.__model__.l1_norm3 = Constraint(rule=self.__l1_rule3(),
                                                 doc='L1 norm 3')

        elif penalty == 2:
            self.__model__.l2_norm = Constraint(rule=self.__l2_rule(),
                                                doc='L2 norm')
        elif penalty == 3:
            self.__model__.lipschitz_norm = Constraint(self.__model__.I,
                                                       rule=self.__lipschitz_rule(),
                                                       doc='Lipschitz norm')
        else:
            raise ValueError('Penalty must be 1, 2, or 3.')

    def __l1_rule1(self):
        """L1 norm #1: right-side equalities"""

        def l1_rule1(model, i, j):
            return model.beta[i, j] <= self.__model__.zeta[i, j]

        return l1_rule1

    def __l1_rule2(self):
        """L1 norm #2: left-side equalities"""

        def l1_rule2(model, i, j):
            return model.beta[i, j] >= -self.__model__.zeta[i, j]

        return l1_rule2

    def __l1_rule3(self):
        """L1 norm #3: sum of zeta"""

        def l1_rule3(model):
            return sum(self.__model__.zeta[ij] for ij in model.I * model.J) == self.eta

        return l1_rule3

    def __l2_rule(self):
        """L2 norm"""

        def l2_rule(model):
            return sum(model.beta[ij] ** 2 for ij in model.I * model.J) <= self.eta

        return l2_rule

    def __lipschitz_rule(self):
        """Lipschitz norm"""

        def lipschitz_rule(model, i):
            return sum(model.beta[i, j] ** 2 for j in model.J) <= self.eta ** 2

        return lipschitz_rule
